import struct
from typing import List

from common.constants import *
from common.errors import *
from common.utils import *
from common.models import *
from common.conn import *

class DataProvider:
	def __init__(self, cluster, donotresolve):
		self.cluster = cluster

		# Caching variables - to prevent many requests to the master:
		self.clusterinfo = None
		self.metaloggers = None
		self.matrices = {}
		self.matrix_summary = None
		self.missing_chunks = None
		self.chunkservers = None
		self.sessions = None
		self.exports = None
		self.sclasses_names = None
		self.sclasses = None
		self.opatterns = None
		self.chunktestinfo = None
		self.openfiles = None
		self.acquiredlocks = None
		self.quotas = None
		self.quotas_maxperc = 0

		self.donotresolve = donotresolve # don't resolve IPs to hostnames
		if self.master()!=None and self.master().has_feature(FEATURE_SESSION_STATS_28):
			self.stats_to_show = 28
		else:
			self.stats_to_show = 16

	def master(self):
		return self.cluster.master()
	
	def get_masterservers(self, IMorder=0, IMrev=False):
		return self.cluster.get_masterservers(IMorder, IMrev)

	def get_metaloggers(self, order=1, rev=False):
		if self.metaloggers==None:
			self.metaloggers=[]
			data,length = self.master().command(CLTOMA_MLOG_LIST,MATOCL_MLOG_LIST)
			if (length%8)==0:
				n = length//8
				for i in range(n):
					d = data[i*8:(i+1)*8]
					v1,v2,v3,ip1,ip2,ip3,ip4 = struct.unpack(">HBBBBBB",d)
					mls = Metalogger((ip1,ip2,ip3,ip4),self.donotresolve,(v1,v2,v3))
					self.metaloggers.append(mls)

		if order is None:
			return self.metaloggers
		
		if   order==1: key=lambda entry: (entry.host)
		elif order==2: key=lambda entry: (entry.sortip)
		elif order==3: key=lambda entry: (entry.sortver)
		else:          key=lambda entry: (entry.host)
		self.metaloggers.sort(key=key)
		if rev:
			self.metaloggers.reverse()
		return self.metaloggers
	
	def get_clusterinfo(self):
		if self.clusterinfo==None:
			masterinfo = self.cluster.masterinfo()
			length = len(masterinfo)
			if length==121 or length==129 or length==137 or length==149 or length==173 or length==181 or length==193 or length==205:
				offset = 8 if (length>=137 and length!=173) else 0
				if offset==8:
					v1,v2,v3,memusage,syscpu,usercpu,totalspace,availspace,freespace,trspace,trfiles,respace,refiles,nodes,dirs,files,chunks = struct.unpack(">HBBQQQQQQQLQLLLLL",masterinfo[:84+offset])
				else:
					v1,v2,v3,memusage,syscpu,usercpu,totalspace,availspace,trspace,trfiles,respace,refiles,nodes,dirs,files,chunks = struct.unpack(">HBBQQQQQQLQLLLLL",masterinfo[:84])
					freespace = None
				if length==205:
					lastsuccessfulstore,lastsaveseconds,lastsavestatus = struct.unpack(">LLB",masterinfo[offset+96:offset+105])
				else:
					lastsuccessfulstore,lastsaveseconds,lastsavestatus = struct.unpack(">LLB",masterinfo[offset+92:offset+101])

				if length>=173:
					metainfomode = 1
					if length==205:
						copychunks,ec8chunks,ec4chunks = struct.unpack(">LLL",masterinfo[offset+84:offset+96])
					else:
						copychunks,ec8chunks = struct.unpack(">LL",masterinfo[offset+84:offset+92])
						ec4chunks = 0
					if length==205:
						chunkcopies,chunkec8parts,chunkec4parts,chunkhypcopies = struct.unpack(">QQQQ",masterinfo[offset+153:offset+185])
					else:
						chunkcopies,chunkec8parts,chunkhypcopies = struct.unpack(">QQQ",masterinfo[offset+149:offset+173])
						chunkec4parts = 0
					allcopies = regularcopies = None
				else:
					metainfomode = 0
					allcopies,regularcopies = struct.unpack(">LL",masterinfo[offset+84:offset+92])
					copychunks = ec8chunks = ec4chunks = None
					chunkcopies = chunkec8parts = chunkec4parts = chunkhypcopies = None
				
				syscpu/=10000000.0
				usercpu/=10000000.0
				lastsaveseconds = lastsaveseconds / 1000.0
			else:
				raise Exception("Unrecognized MATOCL_INFO answer from the Master server (too old version?)")
			
			self.clusterinfo = ClusterInfo((v1,v2,v3),memusage,syscpu,usercpu,totalspace,availspace,freespace,trspace,trfiles,respace,refiles,nodes,dirs,files,chunks,lastsuccessfulstore,lastsaveseconds,lastsavestatus,metainfomode,
																	allcopies,regularcopies,copychunks,ec8chunks,ec4chunks,chunkcopies,chunkec8parts,chunkec4parts,chunkhypcopies)
		return self.clusterinfo
	
	def get_licence(self):
		return None
		
	def get_matrix(self, sclassid=-1):
		if not sclassid in self.matrices or self.progressstatus!=0:
			if self.master().has_feature(FEATURE_SCLASS_IN_MATRIX):
				if sclassid>=0:
					data,length = self.master().command(CLTOMA_CHUNKS_MATRIX,MATOCL_CHUNKS_MATRIX,struct.pack(">BB",0,sclassid))
				else:
					data,length = self.master().command(CLTOMA_CHUNKS_MATRIX,MATOCL_CHUNKS_MATRIX)
			else:
				if sclassid>=0:
					raise RuntimeError("storage class id in matrix is not supported by your master")
				data,length = self.master().command(CLTOMA_CHUNKS_MATRIX,MATOCL_CHUNKS_MATRIX)
			if length==(1+484*2):
				self.progressstatus = struct.unpack(">B",data[0:1])[0]
				self.matrices[sclassid] = ([],[])
				for x in range(2):
					for i in range(11):
						self.matrices[sclassid][x].append(list(struct.unpack(">LLLLLLLLLLL",data[1+x*484+i*44:45+x*484+i*44])))
			elif length==(1+484*6):
				self.progressstatus = struct.unpack(">B",data[0:1])[0]
				self.matrices[sclassid] = ([],[],[],[],[],[])
				for x in range(6):
					for i in range(11):
						self.matrices[sclassid][x].append(list(struct.unpack(">LLLLLLLLLLL",data[1+x*484+i*44:45+x*484+i*44])))
			elif length==(1+484*8):
				self.progressstatus = struct.unpack(">B",data[0:1])[0]
				self.matrices[sclassid] = ([],[],[],[],[],[],[],[])
				for x in range(8):
					for i in range(11):
						self.matrices[sclassid][x].append(list(struct.unpack(">LLLLLLLLLLL",data[1+x*484+i*44:45+x*484+i*44])))
		return (self.matrices[sclassid], self.progressstatus)

	# Calculate sums of missing, undergoal and endangered (for all sclasses)
	# Returns an array of 8: [missing, endangered, undergoal, stable, overgoal, pending deletion, ready to be removed, and total no of chunks
	def get_matrix_summary(self):
		if self.matrix_summary:
			return (self.matrix_summary, self.progressstatus)
		self.matrix_summary = [0] * (MX_COL_TOTAL+1)
		(matrix,_)=self.get_matrix()

		for goal in range(11):
			for actual in range(11):
				(col,_)=redundancy2colclass(goal,actual)
				for i in range(len(matrix)):
					self.matrix_summary[col]+=matrix[0][goal][actual]
					self.matrix_summary[MX_COL_TOTAL]+=matrix[0][goal][actual]
		return (self.matrix_summary, self.progressstatus)

	def get_health_selfcheck(self):
		if (self.master().version_at_least(2,0,66) and self.master().version_less_than(3,0,0)) or self.master().version_at_least(3,0,19):
			data,length = self.master().command(CLTOMA_FSTEST_INFO,MATOCL_FSTEST_INFO,struct.pack(">B",0))
			pver = 1
		else:
			data,length = self.master().command(CLTOMA_FSTEST_INFO,MATOCL_FSTEST_INFO)
			pver = 0
		if length>=(36 + pver*8):
			if pver==1:
				loopstart,loopend,files,ugfiles,mfiles,mtfiles,msfiles,chunks,ugchunks,mchunks,msgbuffleng = struct.unpack(">LLLLLLLLLLL",data[:44])
				datastr = data[44:].decode('utf-8','replace')
			else:
				mtfiles = None
				msfiles = None
				loopstart,loopend,files,ugfiles,mfiles,chunks,ugchunks,mchunks,msgbuffleng = struct.unpack(">LLLLLLLLL",data[:36])
				datastr = data[36:].decode('utf-8','replace')
			return HealthSelfCheck(loopstart,loopend,files,ugfiles,mfiles,mtfiles,msfiles,chunks,ugchunks,mchunks,msgbuffleng,datastr)
		else:
			raise RuntimeError("MFS packet malformed (MATOCL_FSTEST_INFO)")


	# Returns a list of missing chunks (along with files paths) 
	def get_missing_chunks(self, MForder=0, MFrev=False):
		if self.missing_chunks==None:
			self.missing_chunks = []
			missing_inodes = set()
			missings = []
			if (self.master().version_at_least(3,0,25)):
				# get missing chunks list from master
				data,length = self.master().command(CLTOMA_MISSING_CHUNKS,MATOCL_MISSING_CHUNKS,struct.pack(">B",1))
				if length%17==0:
					n = length//17
					for x in range(n):
						chunkid,inode,indx,mtype = struct.unpack(">QLLB",data[x*17:x*17+17])
						missing_inodes.add(inode)
						missings.append((chunkid,inode,indx,mtype))
			inodepaths = self.resolve_inodes_paths(missing_inodes)
			for chunkid,inode,indx,mtype in missings:
				# resolve paths for inode if possible
				if inode in inodepaths:
					paths = inodepaths[inode]
				else:
					paths = []
				self.missing_chunks.append(MissingChunk(paths,inode,indx,chunkid,mtype))

		if MForder is None:
			return self.missing_chunks

		if   MForder==1:  key=lambda mc: (mc.paths)
		elif MForder==2:  key=lambda mc: (mc.inode)
		elif MForder==3:  key=lambda mc: (mc.indx)
		elif MForder==4:  key=lambda mc: (mc.chunkid)
		elif MForder==5:  key=lambda mc: (mc.mtype)
		else:             key=lambda mc: (mc.paths)

		self.missing_chunks.sort(key=key)
		if MFrev:
			self.missing_chunks.reverse()		
		return self.missing_chunks

	# Resolve paths for a list of inodes
	def resolve_inodes_paths(self, inodes):
		def create_chunks(list_name, n):
			listobj = list(list_name)
			for i in range(0, len(listobj), n):
				yield listobj[i:i + n]
		inodepaths = {}
		for chunk in create_chunks(inodes,100):
			if len(chunk)>0:
				data,length = self.master().command(CLTOMA_MASS_RESOLVE_PATHS,MATOCL_MASS_RESOLVE_PATHS,struct.pack(">"+len(chunk)*"L",*chunk))
				pos = 0
				while pos+8<=length:
					inode,psize = struct.unpack(">LL",data[pos:pos+8])
					pos+=8
					if psize == 0:
						if inode not in inodepaths:
							inodepaths[inode] = []
						inodepaths[inode].append("./META")
					elif pos + psize <= length:
						while psize>=4:
							pleng = struct.unpack(">L",data[pos:pos+4])[0]
							pos+=4
							psize-=4
							path = data[pos:pos+pleng]
							pos+=pleng
							psize-=pleng
							path = path.decode('utf-8','replace')
							if inode not in inodepaths:
								inodepaths[inode] = []
							inodepaths[inode].append(path)
						if psize!=0:
							raise RuntimeError("MFS packet malformed (MATOCL_MASS_RESOLVE_PATHS)")
				if pos!=length:
					raise RuntimeError("MFS packet malformed (MATOCL_MASS_RESOLVE_PATHS)")
		return inodepaths

	def get_memory_usage(self):
		data,length = self.master().command(CLTOMA_MEMORY_INFO,MATOCL_MEMORY_INFO)
		if length>=176 and length%16==0:
			memusage = struct.unpack(">QQQQQQQQQQQQQQQQQQQQQQ",data[:176])
			memlabels = ["Chunk hash","Chunks","CS lists","Edge hash","Edges","Node hash","Nodes","Deleted nodes","Chunk tabs","Symlinks","Quota"]
			abrlabels = ["c.h.","c.","c.l.","e.h.","e.","n.h.","n.","d.n.","c.t.","s.","q."]
			totalused = 0
			totalallocated = 0
			for i in range(len(memusage)>>1):
				totalused += memusage[1+i*2]
				totalallocated += memusage[i*2]
			return MemoryUsage(memlabels,abrlabels,totalused,totalallocated,memusage)
		else:
			raise RuntimeError("MFS packet malformed (MATOCL_MEMORY_INFO)")

	# Returns a list of all chunkservers in the cluster
	# CSorder - order of chunkservers in the list, None - no ordering, Default - order by IP address and port nunmber	
	# CSrev - reverse order in the returned list
	# CScsid - return only chunkserver with given id (used for displaying a single chunkserver info)
	def get_chunkservers(self, CSorder=0, CSrev=False):
		if self.chunkservers==None:
			self.chunkservers=[]
			if self.master().has_feature(FEATURE_CSLIST_MODE):
				data,length = self.master().command(CLTOMA_CSERV_LIST,MATOCL_CSERV_LIST,struct.pack(">B",1))
			else:
				data,length = self.master().command(CLTOMA_CSERV_LIST,MATOCL_CSERV_LIST)
			cs = None
			if self.master().version_at_least(4,35,0) and (length%77)==0:
				n = length//77
				for i in range(n):
					d = data[i*77:(i+1)*77]
					flags,v1,v2,v3,oip1,oip2,oip3,oip4,ip1,ip2,ip3,ip4,port,csid,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus,maintenanceto = struct.unpack(">BBBBBBBBBBBBHHQQLQQLLLLLBL",d)
					cs = ChunkServer((oip1,oip2,oip3,oip4),(ip1,ip2,ip3,ip4),self.donotresolve,port,csid,v1,v2,v3,flags,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus,maintenanceto)
					self.chunkservers.append(cs)
			elif self.master().version_at_least(4,26,0) and self.master().version_less_than(4,35,0) and (length%73)==0:
				n = length//73
				for i in range(n):
					d = data[i*73:(i+1)*73]
					flags,v1,v2,v3,oip1,oip2,oip3,oip4,ip1,ip2,ip3,ip4,port,csid,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus = struct.unpack(">BBBBBBBBBBBBHHQQLQQLLLLLB",d)
					cs = ChunkServer((oip1,oip2,oip3,oip4),(ip1,ip2,ip3,ip4),self.donotresolve,port,csid,v1,v2,v3,flags,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus,None)
					self.chunkservers.append(cs)
			elif self.master().version_at_least(3,0,38) and self.master().version_less_than(4,26,0) and (length%69)==0:
				n = length//69
				for i in range(n):
					d = data[i*69:(i+1)*69]
					flags,v1,v2,v3,ip1,ip2,ip3,ip4,port,csid,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus = struct.unpack(">BBBBBBBBHHQQLQQLLLLLB",d)
					cs = ChunkServer((ip1,ip2,ip3,ip4),(ip1,ip2,ip3,ip4),self.donotresolve,port,csid,v1,v2,v3,flags,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus,None)
					self.chunkservers.append(cs)
			elif self.master().version_at_least(2,1,0) and self.master().version_less_than(3,0,38) and (length%68)==0:
				n = length//68
				for i in range(n):
					d = data[i*68:(i+1)*68]
					flags,v1,v2,v3,ip1,ip2,ip3,ip4,port,csid,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels = struct.unpack(">BBBBBBBBHHQQLQQLLLLL",d)
					cs = ChunkServer((ip1,ip2,ip3,ip4),(ip1,ip2,ip3,ip4),self.donotresolve,port,csid,v1,v2,v3,flags,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,None,None)
					self.chunkservers.append(cs)
			else:
				raise RuntimeError("Invalid chunkservers list data (MATOCL_CSERV_LIST)")

		if CSorder is None:
			return self.chunkservers

		if   CSorder==0:  key=lambda cs: (cs.sortip,cs.port)
		elif CSorder==1:  key=lambda cs: (cs.host)
		elif CSorder==2:  key=lambda cs: (cs.sortip,cs.port)
		elif CSorder==3:  key=lambda cs: (cs.port)
		elif CSorder==4:  key=lambda cs: (cs.csid)
		elif CSorder==5:  key=lambda cs: (cs.sortver)
		elif CSorder==6:  key=lambda cs: (cs.gracetime,cs.queue)
		elif CSorder==7:  key=lambda cs: (cs.queue)
		elif CSorder==10: key=lambda cs: (cs.chunks)
		elif CSorder==11: key=lambda cs: (cs.used)
		elif CSorder==12: key=lambda cs: (cs.total)
		elif CSorder==13: key=lambda cs: ((1.0*cs.used)/cs.total if cs.total>0 else 0)
		elif CSorder==20: key=lambda cs: (cs.tdchunks)
		elif CSorder==21: key=lambda cs: (cs.tdused)
		elif CSorder==22: key=lambda cs: (cs.tdtotal)
		elif CSorder==23: key=lambda cs: ((1.0*cs.tdused)/cs.tdtotal if cs.tdtotal>0 else 0)
		else:             key=lambda cs: (0)

		self.chunkservers.sort(key=key)
		if CSrev:
			self.chunkservers.reverse()

		return self.chunkservers

	# Returns a tuple of two lists: first list contains connected chunkservers, 
	# second list contains disconnected chunkservers
	def get_chunkservers_by_state(self, CSorder=0, CSrev=False):
		servers = []
		dservers = []
		for cs in self.get_chunkservers(CSorder, CSrev):
			if cs.is_connected():
				servers.append(cs)
			else:
				dservers.append(cs)
		return servers,dservers


	# Get list of hdds for chunk servers, 
	# HDdata - selects what disks should be returned: "ALL" - all, "ERR" - only with errors, "NOK" - only w/out ok status, "ip:port" - only for a given ip/port chunk server
	# HDorder - sort order
	# HDrev - reverse sort order
	# returns tuple: (hdds,scanhdds)
	def get_hdds(self, HDdata, HDperiod=0, HDtime=0, HDorder=0, HDrev=False):
		# get cs list
		hostlist = []
		multiconn = MFSMultiConn()
		for cs in self.get_chunkservers(None): # don't sort
			if cs.port>0 and cs.is_connected():
				hostip = "%u.%u.%u.%u" % cs.ip
				hostkey = "%s:%u" % (hostip,cs.port)
				sortip = "%03u.%03u.%03u.%03u:%05u" % (cs.ip[0],cs.ip[1],cs.ip[2],cs.ip[3],cs.port)
				if HDdata=="ALL" or HDdata=="ERR" or HDdata=="NOK" or HDdata==hostkey:
					hostlist.append((hostkey,hostip,sortip,cs.port,cs.version,cs.mfrstatus))
					multiconn.register(hostip,cs.port) #register CS to send command to

		if len(hostlist)==0:
			return ([], [])

		# get hdd lists - from all cs at once
		answers=multiconn.command(CLTOCS_HDD_LIST,CSTOCL_HDD_LIST) #ask all registered chunk servers about hdds
		hdds = []
		scanhdds = []
		for hostkey,hostip,sortip,port,version,mfrstatus in hostlist:
			hoststr = resolve(hostip, self.donotresolve)
			if version<=(1,6,8):
				hdds.append((sortip,HDD(CS_HDD_CS_TOO_OLD,hostkey,hoststr,hostip,port,"",sortip,hostkey,hoststr,0,0,0,0,0,0,0,[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],0)))
			elif hostkey not in answers:
				hdds.append((sortip,HDD(CS_HDD_CS_UNREACHABLE,hostkey,hoststr,hostip,port,"",sortip,hostkey,hoststr,0,0,0,0,0,0,0,[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],0)))
			else:
				data,length = answers[hostkey]
				while length>0:
					entrysize = struct.unpack(">H",data[:2])[0]
					entry = data[2:2+entrysize]
					data = data[2+entrysize:]
					length -= 2+entrysize
					plen = entry[0]
					hddpathhex = entry[1:plen+1].hex()
					hddpath = entry[1:plen+1]
					hddpath = hddpath.decode('utf-8','replace')
					hostpath = "%s:%u:%s" % (hoststr,port,hddpath)
					ippath = "%s:%u:%s" % (hostip,port,hddpath)
					clearerrorarg = "%s:%u:%s" % (hostip,port,hddpathhex)
					sortippath = "%s:%s" % (sortip,hddpath)
					flags,errchunkid,errtime,used,total,chunkscnt = struct.unpack(">BQLQQL",entry[plen+1:plen+34])
					rbytes = [0,0,0]
					wbytes = [0,0,0]
					usecreadsum = [0,0,0]
					usecwritesum = [0,0,0]
					usecfsyncsum = [0,0,0]
					rops = [0,0,0]
					wops = [0,0,0]
					fsyncops = [0,0,0]
					usecreadmax = [0,0,0]
					usecwritemax = [0,0,0]
					usecfsyncmax = [0,0,0]
					if entrysize==plen+34+144:
						rbytes[0],wbytes[0],usecreadsum[0],usecwritesum[0],rops[0],wops[0],usecreadmax[0],usecwritemax[0] = struct.unpack(">QQQQLLLL",entry[plen+34:plen+34+48])
						rbytes[1],wbytes[1],usecreadsum[1],usecwritesum[1],rops[1],wops[1],usecreadmax[1],usecwritemax[1] = struct.unpack(">QQQQLLLL",entry[plen+34+48:plen+34+96])
						rbytes[2],wbytes[2],usecreadsum[2],usecwritesum[2],rops[2],wops[2],usecreadmax[2],usecwritemax[2] = struct.unpack(">QQQQLLLL",entry[plen+34+96:plen+34+144])
					elif entrysize==plen+34+192:
						rbytes[0],wbytes[0],usecreadsum[0],usecwritesum[0],usecfsyncsum[0],rops[0],wops[0],fsyncops[0],usecreadmax[0],usecwritemax[0],usecfsyncmax[0] = struct.unpack(">QQQQQLLLLLL",entry[plen+34:plen+34+64])
						rbytes[1],wbytes[1],usecreadsum[1],usecwritesum[1],usecfsyncsum[1],rops[1],wops[1],fsyncops[1],usecreadmax[1],usecwritemax[1],usecfsyncmax[1] = struct.unpack(">QQQQQLLLLLL",entry[plen+34+64:plen+34+128])
						rbytes[2],wbytes[2],usecreadsum[2],usecwritesum[2],usecfsyncsum[2],rops[2],wops[2],fsyncops[2],usecreadmax[2],usecwritemax[2],usecfsyncmax[2] = struct.unpack(">QQQQQLLLLLL",entry[plen+34+128:plen+34+192])
					rbw = [0,0,0]
					wbw = [0,0,0]
					usecreadavg = [0,0,0]
					usecwriteavg = [0,0,0]
					usecfsyncavg = [0,0,0]
					for i in range(3):
						if usecreadsum[i]>0:
							rbw[i] = rbytes[i]*1000000//usecreadsum[i]
						if usecwritesum[i]+usecfsyncsum[i]>0:
							wbw[i] = wbytes[i]*1000000//(usecwritesum[i]+usecfsyncsum[i])
						if rops[i]>0:
							usecreadavg[i] = usecreadsum[i]//rops[i]
						if wops[i]>0:
							usecwriteavg[i] = usecwritesum[i]//wops[i]
						if fsyncops[i]>0:
							usecfsyncavg[i] = usecfsyncsum[i]//fsyncops[i]
					sf = sortippath #default sort field
					if   HDorder==1:	sf = sortippath
					elif HDorder==2:	sf = chunkscnt
					elif HDorder==3:	sf = errtime
					elif HDorder==4:	sf = -flags
					elif HDorder==5:	sf = rbw[HDperiod]
					elif HDorder==6:	sf = wbw[HDperiod]
					elif HDorder==7:	sf = usecreadavg[HDperiod] if HDtime==1 else usecreadmax[HDperiod]
					elif HDorder==8:	sf = usecwriteavg[HDperiod] if HDtime==1 else usecwritemax[HDperiod]
					elif HDorder==9:	sf = usecfsyncavg[HDperiod] if HDtime==1 else usecfsyncmax[HDperiod]
					elif HDorder==10:	sf = rops[HDperiod]
					elif HDorder==11:	sf = wops[HDperiod]
					elif HDorder==12:	sf = fsyncops[HDperiod]
					elif HDorder==20:	sf = used if flags&CS_HDD_SCANNING==0 else 0
					elif HDorder==21:	sf = total if flags&CS_HDD_SCANNING else 0
					elif HDorder==22:	sf = (1.0*used)/total if flags&CS_HDD_SCANNING==0 and total>0 else 0

					if (HDdata=="ERR" and errtime>0) or (HDdata=="NOK" and flags!=0) or (HDdata!="ERR" and HDdata!="NOK"):
						hdd = HDD(CS_HDD_CS_VALID,hostkey,hoststr,hostip,port,hddpath,sortippath,ippath,hostpath,flags,clearerrorarg,errchunkid,errtime,used,total,chunkscnt,rbw,wbw,usecreadavg,usecwriteavg,usecfsyncavg,usecreadmax,usecwritemax,usecfsyncmax,rops,wops,fsyncops,rbytes,wbytes,mfrstatus)
						if flags&CS_HDD_SCANNING: #and not cgimode and ttymode:
							scanhdds.append((sf,hdd))
						else:
							hdds.append((sf,hdd))
		hdds.sort()
		scanhdds.sort()
		if HDrev:
			hdds.reverse()
			scanhdds.reverse()
		return (hdds, scanhdds)
	
	# Summarizes hdd info for given chunkserver,
	# returns tuple: string status of all disks (unknown or ok or warnings or errors) and a list of disks with a few basic fields only
	def cs_hdds_status(self, cs, hdds_all):
		hdds_out = []
		hdds_status = 'ok'
		for (_,hdd) in hdds_all:
			if cs.hostkey!=hdd.hostkey:
				continue
			if (hdd.flags&CS_HDD_INVALID):
				status = 'invalid'
				hdds_status = 'errors'
			elif hdd.flags&CS_HDD_SCANNING:
				status = 'scanning'
			elif (hdd.flags&CS_HDD_DAMAGED) and (hdd.flags&CS_HDD_SCANNING)==0 and (hdd.flags&CS_HDD_INVALID)==0:
				status = 'damaged'
				hdds_status = 'errors'
			elif hdd.flags&CS_HDD_MFR:
				status = 'marked for removal'
			elif hdd.flags==0:
				status = 'ok'
			else:
				status = 'unknown'
				if hdds_status!='errors':
					hdds_status = 'warnings' 
			if hdd.errtime>0:
				status = 'warning' 
				if hdds_status!='errors':
					hdds_status = 'warnings' 
			hdds_out.append({
				'flags': hdd.flags,
				'errtime': hdd.errtime,
				'status': status,
				'used': hdd.used,
				'total': hdd.total
				})
		return (hdds_status, hdds_out)


	
	def get_exports(self, EXorder=0, EXrev=False):
		if self.exports==None:
			self.exports=[]
			if self.master().has_feature(FEATURE_SCLASSGROUPS):
				data,length = self.master().command(CLTOMA_EXPORTS_INFO,MATOCL_EXPORTS_INFO,struct.pack(">B",4))
			elif self.master().has_feature(FEATURE_EXPORT_DISABLES):
				data,length = self.master().command(CLTOMA_EXPORTS_INFO,MATOCL_EXPORTS_INFO,struct.pack(">B",3))
			elif self.master().has_feature(FEATURE_EXPORT_UMASK):
				data,length = self.master().command(CLTOMA_EXPORTS_INFO,MATOCL_EXPORTS_INFO,struct.pack(">B",2))
			else: #if self.master().version_at_least(1,6,26):
				data,length = self.master().command(CLTOMA_EXPORTS_INFO,MATOCL_EXPORTS_INFO,struct.pack(">B",1))
			pos = 0
			while pos<length:
				fip1,fip2,fip3,fip4,tip1,tip2,tip3,tip4,pleng = struct.unpack(">BBBBBBBBL",data[pos:pos+12])
				pos+=12
				path = data[pos:pos+pleng]
				path = path.decode('utf-8','replace')
				pos+=pleng
				if self.master().has_feature(FEATURE_SCLASSGROUPS):
					v1,v2,v3,exportflags,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mintrashretention,maxtrashretention,disables = struct.unpack(">HBBBBHLLLLHLLL",data[pos:pos+38])
					pos+=38
					mingoal = None
					maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				elif self.master().has_feature(FEATURE_EXPORT_DISABLES):
					v1,v2,v3,exportflags,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention,disables = struct.unpack(">HBBBBHLLLLBBLLL",data[pos:pos+38])
					pos+=38
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				elif self.master().has_feature(FEATURE_EXPORT_UMASK):
					v1,v2,v3,exportflags,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention = struct.unpack(">HBBBBHLLLLBBLL",data[pos:pos+34])
					pos+=34
					disables = 0
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				else: #if self.master().version_at_least(1,6,26):
					v1,v2,v3,exportflags,sesflags,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention = struct.unpack(">HBBBBLLLLBBLL",data[pos:pos+32])
					pos+=32
					disables = 0
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
					umaskval = None
				if path=='.':
					meta = 1
					umaskval = None
					disables = 0
				else:
					meta = 0
				expent = ExportsEntry(fip1,fip2,fip3,fip4,tip1,tip2,tip3,tip4,path,meta,v1,v2,v3,exportflags,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mingoal,maxgoal,mintrashretention,maxtrashretention,disables)
				self.exports.append(expent)
		
		if EXorder is None:
			return self.exports
		
		if EXorder==1:  key=lambda ee: (ee.sortipfrom + ee.sortipto)
		elif EXorder==2:  key=lambda ee: (ee.sortipto + ee.sortipfrom)
		elif EXorder==3:  key=lambda ee: (ee.path)
		elif EXorder==4:  key=lambda ee: (ee.sortver)
		elif EXorder==5:  key=lambda ee: (None if ee.meta else ee.is_alldirs())
		elif EXorder==6:  key=lambda ee: (ee.is_password())
		elif EXorder==7:  key=lambda ee: (ee.is_readonly())
		elif EXorder==8:  key=lambda ee: (2-(ee.is_unrestricted()))
		elif EXorder==9:  key=lambda ee: (None if ee.meta else ee.ignore_gid())
		elif EXorder==10: key=lambda ee: (None if ee.meta else ee.is_admin())
		elif EXorder==11: key=lambda ee: (None if ee.meta else ee.rootuid)
		elif EXorder==12: key=lambda ee: (None if ee.meta else ee.rootgid)
		elif EXorder==13: key=lambda ee: (None if ee.meta or (ee.map_user())==0 else ee.mapalluid)
		elif EXorder==14: key=lambda ee: (None if ee.meta or (ee.map_user())==0 else ee.mapallgid)
		elif EXorder==15: key=lambda ee: (ee.sclassgroups)
		elif EXorder==16: key=lambda ee: (ee.sclassgroups)
		elif EXorder==17: key=lambda ee: (ee.mintrashretention)
		elif EXorder==18: key=lambda ee: (ee.maxtrashretention)
		elif EXorder==19: key=lambda ee: (ee.umaskval)
		elif EXorder==20: key=lambda ee: (ee.disables)
		else: key=lambda ee: (ee.sortipfrom + ee.sortipto)

		self.exports.sort(key=key)
		if EXrev:
			self.exports.reverse()

		return self.exports
	
	# Returns a list of all sessions (clients) connected to the cluster
	# MSorder - order of master servers in the list, None - no ordering, default 0 - order by IP
	# MSrev - reverse order in the returned list
	# Returns a list of Session objects
	def get_sessions(self, MSorder=0, MSrev=False):
		if self.sessions==None:
			self.sessions=[]
			if self.master().has_feature(FEATURE_SCLASSGROUPS):
				data,length = self.master().command(CLTOMA_SESSION_LIST,MATOCL_SESSION_LIST,struct.pack(">B",5))
			elif self.master().has_feature(FEATURE_EXPORT_DISABLES):
				data,length = self.master().command(CLTOMA_SESSION_LIST,MATOCL_SESSION_LIST,struct.pack(">B",4))
			elif self.master().has_feature(FEATURE_EXPORT_UMASK):
				data,length = self.master().command(CLTOMA_SESSION_LIST,MATOCL_SESSION_LIST,struct.pack(">B",3))
			else: #if self.master().version_at_least(1,7,8):
				data,length = self.master().command(CLTOMA_SESSION_LIST,MATOCL_SESSION_LIST,struct.pack(">B",2))
			statscnt = struct.unpack(">H",data[0:2])[0]
			pos = 2
			while pos<length:
				sessionid,ip1,ip2,ip3,ip4,v1,v2,v3,openfiles,nsocks,expire,ileng = struct.unpack(">LBBBBHBBLBLL",data[pos:pos+25])
				pos+=25
				info = data[pos:pos+ileng]
				pos+=ileng
				pleng = struct.unpack(">L",data[pos:pos+4])[0]
				pos+=4
				path = data[pos:pos+pleng]
				pos+=pleng
				info = info.decode('utf-8','replace')
				path = path.decode('utf-8','replace')
				if self.master().has_feature(FEATURE_SCLASSGROUPS):
					sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mintrashretention,maxtrashretention,disables = struct.unpack(">BHLLLLHLLL",data[pos:pos+33])
					pos+=33
					mingoal = None
					maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				elif self.master().has_feature(FEATURE_EXPORT_DISABLES):
					sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention,disables = struct.unpack(">BHLLLLBBLLL",data[pos:pos+33])
					pos+=33
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				elif self.master().has_feature(FEATURE_EXPORT_UMASK):
					sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention = struct.unpack(">BHLLLLBBLL",data[pos:pos+29])
					pos+=29
					disables = 0
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
				else: #if self.master().version_at_least(1,6,26):
					sesflags,rootuid,rootgid,mapalluid,mapallgid,mingoal,maxgoal,mintrashretention,maxtrashretention = struct.unpack(">BLLLLBBLL",data[pos:pos+27])
					pos+=27
					disables = 0
					sclassgroups = None
					if mingoal<=1 and maxgoal>=9:
						mingoal = None
						maxgoal = None
					if mintrashretention==0 and maxtrashretention==0xFFFFFFFF:
						mintrashretention = None
						maxtrashretention = None
					umaskval = None
				if statscnt<self.stats_to_show:
					stats_c = struct.unpack(">"+"L"*statscnt,data[pos:pos+4*statscnt])+(0,)*(self.stats_to_show-statscnt)
					pos+=statscnt*4
					stats_l = struct.unpack(">"+"L"*statscnt,data[pos:pos+4*statscnt])+(0,)*(self.stats_to_show-statscnt)
					pos+=statscnt*4
				else:
					stats_c = struct.unpack(">"+"L"*self.stats_to_show,data[pos:pos+4*self.stats_to_show])
					pos+=statscnt*4
					stats_l = struct.unpack(">"+"L"*self.stats_to_show,data[pos:pos+4*self.stats_to_show])
					pos+=statscnt*4
				if path=='.':
					meta=1
				else:
					meta=0
				ses = Session(sessionid,(ip1,ip2,ip3,ip4),self.donotresolve,info,openfiles,nsocks,expire,v1,v2,v3,meta,path,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mingoal,maxgoal,mintrashretention,maxtrashretention,disables,stats_c,stats_l)
				self.sessions.append(ses)

		if MSorder is None:
			return self.sessions

		if   MSorder==1:	key=lambda ses: (ses.sessionid)
		elif MSorder==2:	key=lambda ses: (ses.host)
		elif MSorder==3:	key=lambda ses: (ses.sortip)
		elif MSorder==4:	key=lambda ses: (ses.info)
		elif MSorder==5:	key=lambda ses: (ses.openfiles)
		elif MSorder==6:	key=lambda ses: (ses.nsocks if ses.nsocks>0 else ses.expire)
		elif MSorder==7:	key=lambda ses: (ses.sortver)
		elif MSorder==8:	key=lambda ses: (ses.path)
		elif MSorder==9:	key=lambda ses: (ses.sesflags&1)
		elif MSorder==10:	key=lambda ses: (2-(ses.sesflags&2))
		elif MSorder==11:	key=lambda ses: (-1 if ses.meta else ses.sesflags&4)
		elif MSorder==12:	key=lambda ses: (-1 if ses.meta else ses.sesflags&8)
		elif MSorder==13:	key=lambda ses: (-1 if ses.meta else ses.rootuid)
		elif MSorder==14:	key=lambda ses: (-1 if ses.meta else ses.rootgid)
		elif MSorder==15:	key=lambda ses: (-1 if ses.meta or (ses.sesflags&16)==0 else ses.mapalluid)
		elif MSorder==16:	key=lambda ses: (-1 if ses.meta or (ses.sesflags&16)==0 else ses.mapallgid)
		elif MSorder==17:	key=lambda ses: (ses.sclassgroups)
		elif MSorder==18:	key=lambda ses: (ses.sclassgroups)
		elif MSorder==19:	key=lambda ses: (ses.mintrashretention)
		elif MSorder==20:	key=lambda ses: (ses.maxtrashretention)
		elif MSorder==21:	key=lambda ses: (ses.umaskval)
		elif MSorder==22:	key=lambda ses: (ses.disables)
		else:             key=lambda ses: (ses.sortip)

		self.sessions.sort(key=key)
		if MSrev:
			self.sessions.reverse()
		return self.sessions
	
	# Returns a tuple of two lists: first list contains connected sessions (with nsocks>0), second list contains disconnected sessions with nsocks==0
	def get_sessions_by_state(self, MSorder, MSrev):
		sessions = []
		dsessions = []
		for ses in self.get_sessions(MSorder,MSrev):
			if ses.nsocks>0:
				sessions.append(ses)
			else:
				dsessions.append(ses)
		return sessions,dsessions
	
	# Returns a list of sessions ordered by MO (operations) related parameters
	def get_sessions_order_by_mo(self, MOorder=1, MOrev=False, MOdata=0):
		msessions = []
		for ses in self.get_sessions(None): #don't sort
			if ses.path!='.':
				msessions.append(ses)

		if MOorder is None:
			return msessions

		if   MOorder == 1: key=lambda ses: (ses.host)
		elif MOorder == 2: key=lambda ses: (ses.sortip)
		elif MOorder == 3: key=lambda ses: (ses.info)
		elif MOorder >= 100 and MOorder < 100 + self.stats_to_show:
			if MOdata == 0:
				key=lambda ses: (-ses.stats_l[MOorder-100])
			else:
				key=lambda ses: (-ses.stats_c[MOorder-100])
		elif MOorder == 150:
			if MOdata == 0:
				key=lambda ses: (-sum(ses.stats_l))
			else:
				key=lambda ses: (-sum(ses.stats_c))
		else: key=lambda ses: (ses.host)
		
		msessions.sort(key=key)
		if MOrev:
			msessions.reverse()
		return msessions

	# Returns a list of storage classes names (e.g. for knobs or selects)
	def get_used_sclasses_names(self, order=1):
		if self.sclasses_names==None:
			self.sclasses_names = {}
			data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",128))
			pos = 0
			while pos < length:
				sclassid,sclassnleng = struct.unpack_from(">BB",data,pos)
				pos += 2
				sclassname = data[pos:pos+sclassnleng]
				sclassname = sclassname.decode('utf-8','replace')
				pos += sclassnleng
				has_chunks = struct.unpack_from(">B",data,pos)[0]
				pos += 1
				sclassobj = StorageClassName(sclassname,has_chunks)
				self.sclasses_names[sclassid] = sclassobj
		if order==1:	
			self.sclasses_names = dict(sorted(self.sclasses_names.items()))
		return self.sclasses_names
	
	# Returns a list of storage classes
	def get_sclasses(self, SCorder=0, SCrev=False):
		if self.sclasses==None:
			self.sclasses = []
			show_copy_and_ec = self.master().version_at_least(4,5,0)
			if self.master().has_feature(FEATURE_SCLASSGROUPS):
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",5))
				fver = 5
			elif self.master().has_feature(FEATURE_LABELMODE_OVERRIDES):
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",4))
				fver = 4
			elif self.master().version_at_least(4,34,0):
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",3))
				fver = 3
			elif self.master().version_at_least(4,5,0):
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",2))
				fver = 2
			elif self.master().version_at_least(4,2,0):
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO,struct.pack(">B",1))
				fver = 1
			else:
				data,length = self.master().command(CLTOMA_SCLASS_INFO,MATOCL_SCLASS_INFO)
				fver = 0
			availableservers = struct.unpack(">H",data[:2])[0]
			pos = 2
			while pos < length:
				if show_copy_and_ec:
					createcounters = 6*[None]
					keepcounters = 6*[None]
					archcounters = 6*[None]
					trashcounters = 6*[None]
				else:
					createcounters = 3*[None]
					keepcounters = 3*[None]
					archcounters = 3*[None]
					trashcounters = 3*[None]
				if self.master().version_at_least(4,0,0):
					sclassid,sclassnleng = struct.unpack_from(">BB",data,pos)
					pos += 2
					sclassname = data[pos:pos+sclassnleng]
					sclassname = sclassname.decode('utf-8','replace')
					pos += sclassnleng
					if fver>=5:
						sclassdleng = data[pos]
						pos+=1
						sclassdesc = data[pos:pos+sclassdleng]
						sclassdesc = sclassdesc.decode('utf-8','replace')
						pos += sclassdleng
					else:
						sclassdesc = ""
						sclassdleng = 0
					files,dirs = struct.unpack_from(">LL",data,pos)
					pos += 8
					if fver>=2:
						for cntid in range(6):
							keepcounters[cntid],archcounters[cntid],trashcounters[cntid] = struct.unpack_from(">QQQ",data,pos)
							pos += 3*8
					else:
						for cntid in range(3):
							keepcounters[cntid],archcounters[cntid],trashcounters[cntid] = struct.unpack_from(">QQQ",data,pos)
							pos += 3*8
					if fver>=5:
						priority,export_group = struct.unpack_from(">LB",data,pos)
						pos += 5
					else:
						priority = None
						export_group = sclassid if sclassid<10 else 0
					if fver>=3:
						admin_only,labels_mode,arch_mode,arch_delay,arch_min_size,min_trashretention = struct.unpack_from(">BBBHQH",data,pos)
						pos += 15
					elif fver>=1:
						admin_only,labels_mode,arch_mode,arch_delay,min_trashretention = struct.unpack_from(">BBBHH",data,pos)
						arch_min_size = 0
						pos += 7
					else:
						admin_only,labels_mode,arch_delay,min_trashretention = struct.unpack_from(">BBHH",data,pos)
						arch_mode = 1
						arch_min_size = 0
						pos += 6
					if fver>=4:
						create_canbefulfilled,create_labelscnt,create_uniqmask,create_labelsmode = struct.unpack_from(">BBLB",data,pos)
						pos += 7
						keep_canbefulfilled,keep_labelscnt,keep_uniqmask,keep_labelsmode = struct.unpack_from(">BBLB",data,pos)
						pos += 7
						arch_canbefulfilled,arch_labelscnt,arch_ec_level,arch_uniqmask,arch_labelsmode = struct.unpack_from(">BBBLB",data,pos)
						pos += 8
						trash_canbefulfilled,trash_labelscnt,trash_ec_level,trash_uniqmask,trash_labelsmode = struct.unpack_from(">BBBLB",data,pos)
						pos += 8
					else:
						create_canbefulfilled,create_labelscnt,create_uniqmask = struct.unpack_from(">BBL",data,pos)
						pos += 6
						keep_canbefulfilled,keep_labelscnt,keep_uniqmask = struct.unpack_from(">BBL",data,pos)
						pos += 6
						arch_canbefulfilled,arch_labelscnt,arch_ec_level,arch_uniqmask = struct.unpack_from(">BBBL",data,pos)
						pos += 7
						trash_canbefulfilled,trash_labelscnt,trash_ec_level,trash_uniqmask = struct.unpack_from(">BBBL",data,pos)
						pos += 7
						create_labelsmode = -1
						keep_labelsmode = -1
						arch_labelsmode = -1
						trash_labelsmode = -1
				elif self.master().version_at_least(3,0,75):
					sclassid,sclassnleng = struct.unpack_from(">BB",data,pos)
					pos += 2
					sclassname = data[pos:pos+sclassnleng]
					sclassname = sclassname.decode('utf-8','replace')
					sclassdesc = ""
					sclassdleng = 0
					priority = None
					export_group = sclassid if sclassid<10 else 0
					pos += sclassnleng
					files,dirs,keepcounters[0],archcounters[0],keepcounters[1],archcounters[1],keepcounters[2],archcounters[2],admin_only,labels_mode,arch_delay,create_canbefulfilled,create_labelscnt,keep_canbefulfilled,keep_labelscnt,arch_canbefulfilled,arch_labelscnt = struct.unpack_from(">LLQQQQQQBBHBBBBBB",data,pos)
					pos += 18 + 3 * 16
					if arch_delay==0:
						for cntid in range(3):
							keepcounters[cntid] += archcounters[cntid]
							archcounters[cntid] = None
					min_trashretention = 0
					trash_canbefulfilled = None
					trash_labelscnt = None
					trash_ec_level = 0
					arch_ec_level = 0
					create_uniqmask = 0
					keep_uniqmask = 0
					arch_uniqmask = 0
					trash_uniqmask = 0
					create_labelsmode = -1
					keep_labelsmode = -1
					arch_labelsmode = -1
					trash_labelsmode = -1
					arch_mode = 1
					arch_min_size = 0
				elif self.master().version_at_least(3,0,9):
					sclassid,files,dirs,keepcounters[0],archcounters[0],keepcounters[1],archcounters[1],keepcounters[2],archcounters[2],labels_mode,arch_delay,create_canbefulfilled,create_labelscnt,keep_canbefulfilled,keep_labelscnt,arch_canbefulfilled,arch_labelscnt = struct.unpack_from(">BLLQQQQQQBHBBBBBB",data,pos)
					pos += 18 + 3 * 16
					sclassdesc = ""
					sclassdleng = 0
					priority = None
					export_group = sclassid if sclassid<10 else 0
					admin_only = 0
					if sclassid<10:
						sclassname = str(sclassid)
					else:
						sclassname = "sclass_%u" % (sclassid-9)
					if arch_delay==0:
						for cntid in range(3):
							keepcounters[cntid] += archcounters[cntid]
							archcounters[cntid] = None
					min_trashretention = 0
					trash_canbefulfilled = None
					trash_labelscnt = None
					trash_ec_level = 0
					arch_ec_level = 0
					create_uniqmask = 0
					keep_uniqmask = 0
					arch_uniqmask = 0
					trash_uniqmask = 0
					create_labelsmode = -1
					keep_labelsmode = -1
					arch_labelsmode = -1
					trash_labelsmode = -1
					arch_mode = 1
					arch_min_size = 0
				else:
					sclassid,files,create_canbefulfilled,create_labelscnt = struct.unpack_from(">BLBB",data,pos)
					pos+=7
					sclassdesc = ""
					sclassdleng = 0
					priority = None
					export_group = sclassid if sclassid<10 else 0
					admin_only = 0
					if sclassid<10:
						sclassname = str(sclassid)
					else:
						sclassname = "sclass_%u" % (sclassid-9)
					dirs = 0
					if create_canbefulfilled:
						create_canbefulfilled = 3
					keep_canbefulfilled = create_canbefulfilled
					arch_canbefulfilled = create_canbefulfilled
					keep_labelscnt = create_labelscnt
					arch_labelscnt = create_labelscnt
					labels_mode = 1
					arch_mode = 1
					arch_delay = 0
					arch_min_size = 0
					min_trashretention = 0
					trash_canbefulfilled = None
					trash_labelscnt = None
					trash_ec_level = 0
					arch_ec_level = 0
					create_uniqmask = 0
					keep_uniqmask = 0
					arch_uniqmask = 0
					trash_uniqmask = 0
					create_labelsmode = -1
					keep_labelsmode = -1
					arch_labelsmode = -1
					trash_labelsmode = -1
				if self.master().version_at_least(4,0,0):
					labels_ver = 4
					create_labellist = []
					keep_labellist = []
					arch_labellist = []
					trash_labellist = []
					for i in range(create_labelscnt):
						labelexpr = struct.unpack_from(">"+"B"*SCLASS_EXPR_MAX_SIZE,data,pos)
						pos+=SCLASS_EXPR_MAX_SIZE
						matchingservers = struct.unpack_from(">H",data,pos)[0]
						pos+=2
						create_labellist.append((labelexpr_to_str(labelexpr),matchingservers,labelexpr))
					for i in range(keep_labelscnt):
						labelexpr = struct.unpack_from(">"+"B"*SCLASS_EXPR_MAX_SIZE,data,pos)
						pos+=SCLASS_EXPR_MAX_SIZE
						matchingservers = struct.unpack_from(">H",data,pos)[0]
						pos+=2
						keep_labellist.append((labelexpr_to_str(labelexpr),matchingservers,labelexpr))
					for i in range(arch_labelscnt):
						labelexpr = struct.unpack_from(">"+"B"*SCLASS_EXPR_MAX_SIZE,data,pos)
						pos+=SCLASS_EXPR_MAX_SIZE
						matchingservers = struct.unpack_from(">H",data,pos)[0]
						pos+=2
						arch_labellist.append((labelexpr_to_str(labelexpr),matchingservers,labelexpr))
					for i in range(trash_labelscnt):
						labelexpr = struct.unpack_from(">"+"B"*SCLASS_EXPR_MAX_SIZE,data,pos)
						pos+=SCLASS_EXPR_MAX_SIZE
						matchingservers = struct.unpack_from(">H",data,pos)[0]
						pos+=2
						trash_labellist.append((labelexpr_to_str(labelexpr),matchingservers,labelexpr))
				else:
					labels_ver = 3
					create_labellist = []
					for i in range(create_labelscnt):
						labelmasks = struct.unpack_from(">"+"L"*MASKORGROUP,data,pos)
						pos+=4*MASKORGROUP
						matchingservers = struct.unpack_from(">H",data,pos)[0]
						pos+=2
						create_labellist.append((labelmasks_to_str(labelmasks),matchingservers,labelmasks))
					if self.master().version_at_least(3,0,9):
						keep_labellist = []
						for i in range(keep_labelscnt):
							labelmasks = struct.unpack_from(">"+"L"*MASKORGROUP,data,pos)
							pos+=4*MASKORGROUP
							matchingservers = struct.unpack_from(">H",data,pos)[0]
							pos+=2
							keep_labellist.append((labelmasks_to_str(labelmasks),matchingservers,labelmasks))
						arch_labellist = []
						for i in range(arch_labelscnt):
							labelmasks = struct.unpack_from(">"+"L"*MASKORGROUP,data,pos)
							pos+=4*MASKORGROUP
							matchingservers = struct.unpack_from(">H",data,pos)[0]
							pos+=2
							arch_labellist.append((labelmasks_to_str(labelmasks),matchingservers,labelmasks))
					else:
						keep_labellist = create_labellist
						arch_labellist = create_labellist
					trash_labellist = []
					
				states = []
				if len(create_labellist)>0:
					states.append(StorageClassState(1,"CREATE",createcounters,None,create_labellist,create_uniqmask,create_labelsmode,create_canbefulfilled))
				else:
					states.append(StorageClassState(0,"CREATE",createcounters,None,keep_labellist,keep_uniqmask,keep_labelsmode,keep_canbefulfilled))
				states.append(StorageClassState(1,"KEEP",keepcounters,None,keep_labellist,keep_uniqmask,keep_labelsmode,keep_canbefulfilled))
				if len(arch_labellist)>0 or arch_ec_level>0:
					states.append(StorageClassState(1,"ARCH",archcounters,arch_ec_level,arch_labellist,arch_uniqmask,arch_labelsmode,arch_canbefulfilled))
				else:
					states.append(StorageClassState(0,"ARCH",archcounters,None,keep_labellist,keep_uniqmask,keep_labelsmode,keep_canbefulfilled))
				if len(trash_labellist)>0 or trash_ec_level>0:
					states.append(StorageClassState(1,"TRASH",trashcounters,trash_ec_level,trash_labellist,trash_uniqmask,trash_labelsmode,trash_canbefulfilled))
				else:
					states.append(StorageClassState(0,"TRASH",trashcounters,None,keep_labellist,keep_uniqmask,keep_labelsmode,keep_canbefulfilled))
			
				sc = StorageClass(sclassid,sclassname,sclassdesc,priority,export_group,admin_only,labels_mode,arch_mode,arch_delay,arch_min_size,min_trashretention,files,dirs,states,availableservers,labels_ver)
				self.sclasses.append(sc)

		if SCorder is None:
			return self.sclasses

		if   SCorder==1:  key=lambda sc: (sc.sclassid)
		elif SCorder==2:  key=lambda sc: (sc.sclassname)
		elif SCorder==3:  key=lambda sc: (sc.admin_only)
		elif SCorder==4:  key=lambda sc: (sc.labels_mode)
		elif SCorder==5:  key=lambda sc: (sc.arch_mode)
		elif SCorder==6:  key=lambda sc: (sc.arch_delay)
		elif SCorder==7:  key=lambda sc: (sc.arch_min_size)
		elif SCorder==8:  key=lambda sc: (sc.min_trashretention)
		elif SCorder==9:  key=lambda sc: (sc.files)
		elif SCorder==10: key=lambda sc: (sc.dirs)
		else: key=lambda sc: (sc.sclassname)

		self.sclasses.sort(key=key)
		if SCrev:
			self.sclasses.reverse()

		return self.sclasses

	def get_opatterns(self, PAorder=1, PArev=False):
		if self.opatterns==None:
			self.opatterns = []
			data,length = self.master().command(CLTOMA_PATTERN_INFO,MATOCL_PATTERN_INFO)
			pos = 0
			while pos < length:
				globleng = struct.unpack_from(">B",data,pos)[0]
				pos += 1
				globname = data[pos:pos+globleng]
				globname = globname.decode('utf-8','replace')
				pos += globleng
				euid,egid,priority,omask,sclassleng = struct.unpack_from(">LLBBB",data,pos)
				pos += 11
				sclassname = data[pos:pos+sclassleng]
				sclassname = sclassname.decode('utf-8','replace')
				pos += sclassleng
				trashretention,seteattr,clreattr = struct.unpack_from(">HBB",data,pos)
				if (omask&PATTERN_OMASK_SCLASS)==0:
					sclassname = None
				if (omask&PATTERN_OMASK_TRASHRETENTION)==0:
					trashretention = None
				if (omask&PATTERN_OMASK_EATTR)==0:
					seteattr = 0
					clreattr = 0
				pos += 4
				self.opatterns.append(OPattern(globname,euid,egid,priority,sclassname,trashretention,seteattr,clreattr))

		if PAorder is None:
			return self.opatterns

		if PAorder==1: key=lambda op: (op.globname)
		elif PAorder==2: key=lambda op: (op.euid)
		elif PAorder==3: key=lambda op: (op.egid)
		elif PAorder==4: key=lambda op: (op.priority)
		elif PAorder==5: key=lambda op: (op.sclassname)
		elif PAorder==6: key=lambda op: (op.trashretention)
		elif PAorder==7: key=lambda op: (op.seteattr|op.clreattr)
		else: key=lambda op: (op.globname)

		self.opatterns.sort(key=key)
		if PArev:
			self.opatterns.reverse()
		
		return self.opatterns
	
	def get_chunktestinfo(self):
		if self.chunktestinfo==None:
			data,length = self.master().command(CLTOMA_CHUNKSTEST_INFO,MATOCL_CHUNKSTEST_INFO)
			if length==72:
				self.chunktestinfo = ChunkTestInfo72(*struct.unpack(">LLLLLLLLLLLLLLLLLL",data))
			elif length==96:
				self.chunktestinfo = ChunkTestInfo96(*struct.unpack(">"+24*"L",data))
			else:
				raise Exception("Invalid length of chunktestinfo response")
		return self.chunktestinfo
	
	# returns a list of open files for a given session, if sessionid is 0, returns a list of all open files
	def get_openfiles(self, OFsessionid=0, OForder=1, OFrev=False):
		if self.openfiles==None:
			self.openfiles = []
			sessionsdata = {}
			for ses in self.get_sessions():
				if ses.sessionid>0 and ses.sessionid < 0x80000000:
					sessionsdata[ses.sessionid]=(ses.host,ses.sortip,ses.strip,ses.info,ses.openfiles)
			inodes = set()
			data,length = self.master().command(CLTOMA_LIST_OPEN_FILES,MATOCL_LIST_OPEN_FILES,struct.pack(">L",OFsessionid))
			openfiles = []
			if OFsessionid==0:
				n = length//8
				for x in range(n):
					sessionid,inode = struct.unpack(">LL",data[x*8:x*8+8])
					openfiles.append((sessionid,inode))
					inodes.add(inode)
			else:
				n = length//4
				for x in range(n):
					inode = struct.unpack(">L",data[x*4:x*4+4])[0]
					openfiles.append((OFsessionid,inode))
					inodes.add(inode)
			inodepaths = self.resolve_inodes_paths(inodes)

			for sessionid,inode in openfiles:
				if sessionid in sessionsdata:
					host,sortipnum,ipnum,info,openfiles = sessionsdata[sessionid]
				else:
					host = 'unknown'
					sortipnum = ''
					ipnum = ''
					info = 'unknown'
				if inode in inodepaths:
					paths = inodepaths[inode]
				else:
					paths = []
				self.openfiles.append(OpenFile(sessionid,host,sortipnum,ipnum,info,inode,paths))

		if OForder is None:
			return self.openfiles

		if   OForder==1: key=lambda of: (of.sessionid)
		elif OForder==2: key=lambda of: (of.host)
		elif OForder==3: key=lambda of: (of.sortipnum)
		elif OForder==4: key=lambda of: (of.info)
		elif OForder==5: key=lambda of: (of.inode)
		elif OForder==6: key=lambda of: (of.paths)
		else:            key=lambda of: (of.sortipnum)

		self.openfiles.sort(key=key)
		if OFrev:
			self.openfiles.reverse()
		
		return self.openfiles
	
	def get_acquiredlocks(self, ALinode=0, ALorder=1, ALrev=False):
		if self.acquiredlocks==None:
			self.acquiredlocks = []
			sessionsdata = {}
			for ses in self.get_sessions():
				if ses.sessionid>0 and ses.sessionid < 0x80000000:
					sessionsdata[ses.sessionid]=(ses.host,ses.sortip,ses.strip,ses.info)

			data,length = self.master().command(CLTOMA_LIST_ACQUIRED_LOCKS,MATOCL_LIST_ACQUIRED_LOCKS,struct.pack(">L",ALinode))
			locks = []
			if ALinode==0:
				n = length//37
				for x in range(n):
					inode,sessionid,owner,pid,start,end,ctype = struct.unpack(">LLQLQQB",data[x*37:x*37+37])
					locks.append((inode,sessionid,owner,pid,start,end,ctype))
			else:
				n = length//33
				for x in range(n):
					sessionid,owner,pid,start,end,ctype = struct.unpack(">LQLQQB",data[x*33:x*33+33])
					locks.append((ALinode,sessionid,owner,pid,start,end,ctype))
			for inode,sessionid,owner,pid,start,end,ctype in locks:
				if sessionid in sessionsdata:
					host,sortipnum,ipnum,info = sessionsdata[sessionid]
				else:
					host = 'unknown'
					sortipnum = ''
					ipnum = ''
					info = 'unknown'
				self.acquiredlocks.append(AcquiredLock(inode,sessionid,host,sortipnum,ipnum,info,owner,pid,start,end,ctype))

		if ALorder is None:
			return self.acquiredlocks

		if   ALorder==1: key=lambda al: (al.inode)
		elif ALorder==2: key=lambda al: (al.sessionid)
		elif ALorder==3: key=lambda al: (al.host)
		elif ALorder==4: key=lambda al: (al.sortipnum)
		elif ALorder==5: key=lambda al: (al.info)
		elif ALorder==6: key=lambda al: (al.locktype)
		elif ALorder==7: key=lambda al: (al.owner)
		elif ALorder==8: key=lambda al: (al.pid)
		elif ALorder==9: key=lambda al: (al.start)
		elif ALorder==10: key=lambda al: (al.end)
		elif ALorder==11: key=lambda al: (al.ctype)
		else:            key=lambda of: (of.inode)

		self.acquiredlocks.sort(key=key)
		if ALrev:
			self.acquiredlocks.reverse()
		
		return self.acquiredlocks		

	def get_quotas(self, QUorder=1, QUrev=False):
		if self.quotas==None:
			self.quotas = []
			self.quotas_maxperc = 0.0
			if self.master().has_feature(FEATURE_DEFAULT_GRACEPERIOD):
				data,length = self.master().command(CLTOMA_QUOTA_INFO,MATOCL_QUOTA_INFO,struct.pack(">B",1))
			else:
				data,length = self.master().command(CLTOMA_QUOTA_INFO,MATOCL_QUOTA_INFO)
			if length>=4:
				maxperc = 0.0
				pos = 0
				while pos<length:
					inode,pleng = struct.unpack(">LL",data[pos:pos+8])
					pos+=8
					path = data[pos:pos+pleng]
					path = path.decode('utf-8','replace')
					pos+=pleng
					if self.master().version_at_least(3,0,9):
						graceperiod,exceeded,qflags,timetoblock = struct.unpack(">LBBL",data[pos:pos+10])
						pos+=10
					else:
						exceeded,qflags,timetoblock = struct.unpack(">BBL",data[pos:pos+6])
						pos+=6
						graceperiod = 0
					sinodes,slength,ssize,srealsize = struct.unpack(">LQQQ",data[pos:pos+28])
					pos+=28
					hinodes,hlength,hsize,hrealsize = struct.unpack(">LQQQ",data[pos:pos+28])
					pos+=28
					cinodes,clength,csize,crealsize = struct.unpack(">LQQQ",data[pos:pos+28])
					pos+=28
					if (qflags&1) and sinodes>0:
						perc = 100.0*cinodes/sinodes
						if perc>maxperc: maxperc = perc
					if (qflags&2) and slength>0:
						perc = 100.0*clength/slength
						if perc>maxperc: maxperc = perc
					if (qflags&4) and ssize>0:
						perc = 100.0*csize/ssize
						if perc>maxperc: maxperc = perc
					if (qflags&8) and srealsize>0:
						perc = 100.0*crealsize/srealsize
						if perc>maxperc: maxperc = perc
					if (qflags&16) and hinodes>0:
						perc = 100.0*cinodes/hinodes
						if perc>maxperc: maxperc = perc
					if (qflags&32) and hlength>0:
						perc = 100.0*clength/hlength
						if perc>maxperc: maxperc = perc
					if (qflags&64) and hsize>0:
						perc = 100.0*csize/hsize
						if perc>maxperc: maxperc = perc
					if (qflags&128) and hrealsize>0:
						perc = 100.0*crealsize/hrealsize
						if perc>maxperc: maxperc = perc
					self.quotas.append(Quota(path,exceeded,qflags,graceperiod,timetoblock,sinodes,slength,ssize,srealsize,hinodes,hlength,hsize,hrealsize,cinodes,clength,csize,crealsize))
				self.quotas_maxperc = maxperc + 0.01

		if QUorder is None:
			return self.quotas_maxperc,self.quotas
				
		if   QUorder==1:  key=lambda qu: (qu.path)
		elif QUorder==2:  key=lambda qu: (qu.exceeded)
		elif QUorder==9:  key=lambda qu: (qu.graceperiod)
		elif QUorder==10: key=lambda qu: (qu.timetoblock)
		elif QUorder==11: key=lambda qu: (qu.sinodes)
		elif QUorder==12: key=lambda qu: (qu.slength)
		elif QUorder==13: key=lambda qu: (qu.ssize)
		elif QUorder==14: key=lambda qu: (qu.srealsize)
		elif QUorder==21: key=lambda qu: (qu.hinodes)
		elif QUorder==22: key=lambda qu: (qu.hlength)
		elif QUorder==23: key=lambda qu: (qu.hsize)
		elif QUorder==24: key=lambda qu: (qu.hrealsize)
		elif QUorder==31: key=lambda qu: (qu.cinodes)
		elif QUorder==32: key=lambda qu: (qu.clength)
		elif QUorder==33: key=lambda qu: (qu.csize)
		elif QUorder==34: key=lambda qu: (qu.crealsize)
		elif QUorder==41: key=lambda qu: ((-1,0) if (qu.qflags&1)==0 else (1,0) if qu.sinodes==0 else (0,1.0*qu.cinodes/qu.sinodes))
		elif QUorder==42: key=lambda qu: ((-1,0) if (qu.qflags&2)==0 else (1,0) if qu.slength==0 else (0,1.0*qu.clength/qu.slength))
		elif QUorder==43: key=lambda qu: ((-1,0) if (qu.qflags&4)==0 else (1,0) if qu.ssize==0 else (0,1.0*qu.csize/qu.ssize))
		elif QUorder==44: key=lambda qu: ((-1,0) if (qu.qflags&8)==0 else (1,0) if qu.srealsize==0 else (0,1.0*qu.crealsize/qu.srealsize))
		elif QUorder==51: key=lambda qu: ((-1,0) if (qu.qflags&16)==0 else (1,0) if qu.hinodes==0 else (0,1.0*qu.cinodes/qu.hinodes))
		elif QUorder==52: key=lambda qu: ((-1,0) if (qu.qflags&32)==0 else (1,0) if qu.hlength==0 else (0,1.0*qu.clength/qu.hlength))
		elif QUorder==53: key=lambda qu: ((-1,0) if (qu.qflags&64)==0 else (1,0) if qu.hsize==0 else (0,1.0*qu.csize/qu.hsize))
		elif QUorder==54: key=lambda qu: ((-1,0) if (qu.qflags&128)==0 else (1,0) if qu.hrealsize==0 else (0,1.0*qu.crealsize/qu.hrealsize))
		else:             key=lambda of: (of.path)

		self.quotas.sort(key=key)
		if QUrev:
			self.quotas.reverse()
		
		return self.quotas_maxperc,self.quotas
	



from common.models import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	out = []
	ci = dp.get_chunktestinfo()
	if type(ci)==ChunkTestInfo72:
		out.append("""<div class="tab_title">Chunks housekeeping operation</div>""")
		out.append("""<table class="FR no-hover">""")
		out.append("""	<tr>""")
		out.append("""		<th colspan="2">Loop time</th>""")
		out.append("""		<th colspan="4">Deletions</th>""")
		out.append("""		<th colspan="3">Replications</th>""")
		out.append("""		<th colspan="2">Locked</th>""")
		out.append("""	</tr>""")
		out.append("""	<tr>""")
		out.append("""		<th>start</th>""")
		out.append("""		<th>end</th>""")
		out.append("""		<th>invalid</th>""")
		out.append("""		<th>unused</th>""")
		out.append("""		<th>disk clean</th>""")
		out.append("""		<th>over goal</th>""")
		out.append("""		<th>under goal</th>""")
		out.append("""		<th>wrong labels</th>""")
		out.append("""		<th>rebalance</th>""")
		out.append("""		<th>unused</th>""")
		out.append("""		<th>used</th>""")
		out.append("""	</tr>""")
		if ci.loopstart>0:
			out.append("""	<tr>""")
			out.append("""		<td align="center">%s</td>""" % (time.asctime(time.localtime(ci.loopstart)),))
			out.append("""		<td align="center">%s</td>""" % (time.asctime(time.localtime(ci.loopend)),))
			out.append("""		<td align="right">%u/%u</td>""" % (ci.del_invalid,ci.del_invalid+ci.ndel_invalid))
			out.append("""		<td align="right">%u/%u</td>""" % (ci.del_unused,ci.del_unused+ci.ndel_unused))
			out.append("""		<td align="right">%u/%u</td>""" % (ci.del_dclean,ci.del_dclean+ci.ndel_dclean))
			out.append("""		<td align="right">%u/%u</td>""" % (ci.del_ogoal,ci.del_ogoal+ci.ndel_ogoal))
			out.append("""		<td align="right">%u/%u</td>""" % (ci.rep_ugoal,ci.rep_ugoal+ci.nrep_ugoal))
			out.append("""		<td align="right">%u/%u/%u</td>""" % (ci.rep_wlab,ci.labels_dont_match,ci.rep_wlab+ci.nrep_wlab+ci.labels_dont_match))
			out.append("""		<td align="right">%u</td>""" % ci.rebalance)
			out.append("""		<td align="right">%u</td>""" % ci.locked_unused)
			out.append("""		<td align="right">%u</td>""" % ci.locked_used)
			out.append("""	</tr>""")
		else:
			out.append("""	<tr>""")
			out.append("""		<td colspan="11" align="center">No data, self-check loop not finished yet</td>""")
			out.append("""	</tr>""")
		out.append("""</table>""")

	elif type(ci)==ChunkTestInfo96:
		if ci.loopstart>0:
			out.append("""<div class="tab_title">Chunks housekeeping operations</div>""")
			out.append("""<table class="FR panel no-hover">""")
			out.append(""" <tr><td>""")
			tdata=[["Loop start", datetime_to_str(ci.loopstart)], 
					["Loop end", datetime_to_str(ci.loopend)], 
					["Locked unused chunks",decimal_number_html(ci.locked_unused)],
					["Locked chunks", decimal_number_html(ci.locked_used)],
					["Fixed chunks", decimal_number_html(ci.fixed)],
					["Forced keep mode", decimal_number_html(ci.forcekeep)]]
			out.append(html_table_vertical("Housekeeping loop statistics", tdata))
			out.append(""" </td><td>""")
			tdata=[["Invalid", decimal_number_html(ci.delete_invalid)], 
					["Removed", decimal_number_html(ci.delete_no_longer_needed)], 
					["Wrong version", decimal_number_html(ci.delete_wrong_version)],
					["Excess", decimal_number_html(ci.delete_excess_copy)],
					["Marked for removal", decimal_number_html(ci.delete_diskclean_copy)]]
			out.append(html_table_vertical("Chunk copies - deletions", tdata))

			tdata=[["Needed", decimal_number_html(ci.replicate_needed_copy)], 
					["Wrong labels", decimal_number_html(ci.replicate_wronglabels_copy)]]
			out.append(html_table_vertical("Chunks copies - replications", tdata))	
			out.append(""" </td><td>""")
			tdata=[["Duplicated", decimal_number_html(ci.delete_duplicated_ecpart)], 
					["Excess", decimal_number_html(ci.delete_excess_ecpart)], 
					["Marked for removal", decimal_number_html(ci.delete_diskclean_ecpart)]]
			out.append(html_table_vertical("EC parts - deletions", tdata))

			tdata=[["Duplicated server", decimal_number_html(ci.replicate_dupserver_ecpart)], 
					["Needed", decimal_number_html(ci.replicate_needed_ecpart)], 
					["Wrong labels", decimal_number_html(ci.replicate_wronglabels_ecpart)],
					["Recovered", decimal_number_html(ci.recover_ecpart)],
					["Calculated checksums", decimal_number_html(ci.calculate_ecchksum)]]
			out.append(html_table_vertical("EC parts - replications", tdata))
			out.append(""" </td><td>""")
			tdata=[["Split: copies &rarr; EC parts", decimal_number_html(ci.split_copy_into_ecparts)], 
					["Join: EC parts &rarr; copies", decimal_number_html(ci.join_ecparts_into_copy)]]
			out.append(html_table_vertical("Copies &harr; EC parts", tdata))	

			tdata=[["Rebalance", decimal_number_html(ci.replicate_rebalance)]]
			out.append(html_table_vertical("Replications", tdata))	

			out.append(""" </td></tr>""")
			out.append("""</table>""")						
		else:
			out.append("""<div class="tab_title">Chunks housekeeping operations</div>""")
			out.append("""<table class="FR">""")
			out.append("""	<tr>""")
			out.append("""		<td align="center">No data, self-check loop not finished yet</td>""")
			out.append("""	</tr>""")
			out.append("""</table>""")
	return out



from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	EXorder = fields.getint("EXorder", 0)
	EXrev = fields.getint("EXrev", 0)

	out = []
	out.append("""<div class="tab_title">Exports</div>""")
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfsexports">""")
	out.append("""	<tr>""")
	out.append("""		<th rowspan="2" class="acid_tab_enumerate">#</th>""")
	out.append("""		<th colspan="2">IP&nbsp;range</th>""")
	out.append("""		<th rowspan="2">Path</th>""")
	out.append("""		<th rowspan="2">Minimum<br/>version</th>""")
	out.append("""		<th rowspan="2">Alldirs</th>""")
	out.append("""		<th rowspan="2">Password</th>""")
	out.append("""		<th rowspan="2">RO/RW</th>""")
	out.append("""		<th rowspan="2">Restricted&nbsp;IP</th>""")
	out.append("""		<th rowspan="2">Ignore<br/>gid</th>""")
	out.append("""		<th rowspan="2">Admin</th>""")
	out.append("""		<th colspan="2">Map&nbsp;root</th>""")
	out.append("""		<th colspan="2">Map&nbsp;users</th>""")
	out.append("""		<th rowspan="2">Allowed<br/>sclasses</th>""")
	out.append("""		<th colspan="2">Trashretention&nbsp;limit</th>""")
	if dp.master().has_feature(FEATURE_EXPORT_UMASK):
		out.append("""		<th rowspan="2">Global<br/>umask</th>""")
	if dp.master().has_feature(FEATURE_EXPORT_DISABLES):
		out.append("""		<th rowspan="2">Disables<br/>mask</th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<th style="min-width:80px;">from</th>""")
	out.append("""		<th style="min-width:80px;">to</th>""")
	out.append("""		<th style="min-width:40px;">uid</th>""")
	out.append("""		<th style="min-width:40px;">gid</th>""")
	out.append("""		<th style="min-width:40px;">uid</th>""")
	out.append("""		<th style="min-width:40px;">gid</th>""")
	out.append("""		<th style="min-width:40px;">min</th>""")
	out.append("""		<th style="min-width:40px;">max</th>""")
	out.append("""	</tr>""")

	for ee in dp.get_exports(EXorder,EXrev):
		out.append("""	<tr>""")
		out.append("""		<td align="right"></td>""")
		out.append("""		<td align="center"><span class="sortkey">%s</span>%s</td>""" % (ee.sortipfrom,ee.stripfrom))
		out.append("""		<td align="center"><span class="sortkey">%s</span>%s</td>""" % (ee.sortipto,ee.stripto))
		out.append("""		<td align="left">%s</td>""" % (".&nbsp;(META)" if ee.meta else htmlentities(ee.path)))
		out.append("""		<td align="center"><span class="sortkey">%s</span>%s</td>""" % (ee.sortver,ee.strver))
		out.append("""		<td align="center">%s</td>""" % ("-" if ee.meta else "yes" if ee.is_alldirs() else "no"))
		out.append("""		<td align="center">%s</td>""" % ("yes" if ee.is_password() else "no"))
		out.append("""		<td align="center">%s</td>""" % ("ro" if ee.is_readonly() else "rw"))
		out.append("""		<td align="center">%s</td>""" % ("no" if ee.is_unrestricted() else "yes"))
		out.append("""		<td align="center">%s</td>""" % ("-" if ee.meta else "yes" if ee.ignore_gid() else "no"))
		out.append("""		<td align="center">%s</td>""" % ("-" if ee.meta else "yes" if ee.is_admin() else "no"))
		if ee.meta:
			out.append("""		<td align="right"><span class="sortkey">-1</span>-</td>""")
			out.append("""		<td align="right"><span class="sortkey">-1</span>-</td>""")
		else:
			out.append("""		<td align="right">%u</td>""" % ee.rootuid)
			out.append("""		<td align="right">%u</td>""" % ee.rootgid)
		if ee.meta or (ee.map_user())==0:
			out.append("""		<td align="center"><span class="sortkey">-1</span>&nbsp;</td>""")
			out.append("""		<td align="center"><span class="sortkey">-1</span>&nbsp;</td>""")
		else:
			out.append("""		<td align="right">%u</td>""" % ee.mapalluid)
			out.append("""		<td align="right">%u</td>""" % ee.mapallgid)
		out.append("""		<td align="center"><span class="sortkey">%u</span>%s</td>""" % (ee.sclassgroups_sort,ee.sclassgroups_str))
		if ee.mintrashretention!=None and ee.maxtrashretention!=None:
			out.append("""		<td align="right"><span class="sortkey">%u</span><span data-tt-info="%s">%s</span></td>""" % (ee.mintrashretention,timeduration_to_fullstr(ee.mintrashretention),timeduration_to_shortstr(ee.mintrashretention)))
			out.append("""		<td align="right"><span class="sortkey">%u</span><span data-tt-info="%s">%s</span></td>""" % (ee.maxtrashretention,timeduration_to_fullstr(ee.maxtrashretention),timeduration_to_shortstr(ee.maxtrashretention)))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1</span>&nbsp;</td>""")
			out.append("""		<td align="center"><span class="sortkey">-1</span>&nbsp;</td>""")
		if dp.master().has_feature(FEATURE_EXPORT_UMASK):
			if ee.umaskval==None:
				out.append("""		<td align="center"><span class="sortkey">-1</span>-</td>""")
			else:
				out.append("""		<td align="center">%03o</td>""" % ee.umaskval)
		if dp.master().has_feature(FEATURE_EXPORT_DISABLES):
			if ee.disables!=0:
				helpmsg = "Disabled operations:<br/>%s" % disablesmask_to_string(ee.disables)
			else:
				helpmsg = "All operations enabled"
			out.append("""		<td align="center"><span class="sortkey">%u</span><span data-tt-info="%s">%08X</span></td>""" % (ee.disables,helpmsg,ee.disables))
		out.append("""	</tr>""")

	out.append("""</table>""")
	return out


import os
import json

from common.constants import *

# Contains 0, 1 or many issues (errors and/or warnings and/or infos) found by Validator's check_* routine
class Issues:
	# Creates a new Issues object, if 'issue' is provided this is added (initiated) to the list of issues
	def __init__(self, validator, issue=None, css_class_override=None, icon_override=None):
		self.issues_dict = validator.issues_dict
		self.issues = []
		self.css_class_override = css_class_override
		self.icon_override = icon_override
		self.has_error = False
		self.has_warning = False
		self.has_info = False
		self.has_help = False
		if issue:
			self.add_issue(issue)

	def __iter__(self):
		return iter(self.issues)

	def len(self):
		return len(self.issues)
	def any(self):
		return len(self.issues)>0
	
	# Appends an issue with a given id to the returned list of issues; severity of the issue is determined automatically based on issues_dict ('help.json')
	def add_issue(self, id):
		def find_issue_by_id(id):
			if id in self.issues_dict:
				return self.issues_dict[id]  # Return the first matching dictionary
			else:
				return None  # Return None if no match is found
		issue = find_issue_by_id(id)
		if issue == None:
			self.css_class_override = "missing-data-tt"
			self.issues.append('unknown_severity')
			self.issues.append(id)
			return

		self.issues.append(id)

		if issue['severity']=="Error":
			self.has_error = True
		elif issue['severity']=="Warning":
			self.has_warning = True
		elif issue['severity']=="Info":
			self.has_info = True
		elif issue['severity']=="Help":
			self.has_help = True

	# Returns a list of issues ids as comma separated string for use directly in the 'data-tt' attribute
	def data_help(self):
		return ",".join(self.issues)
	
	# A convenient helper: returns a string with a complete <span> tag containing class and data-tt attributes,
	# along with icon and a given msg string as its inner html
	def span(self, inner_html='', clz=''):
		if clz:
			if self.css_class():
				clz='%s %s' % (clz, self.css_class())
		elif self.css_class():
			clz = self.css_class()
		if clz:
			clz = ' class="%s"' % clz
		else:
			clz=''
		if self.data_help():
			data_help = ' data-tt="%s"' % self.data_help()
		else:
			data_help = ''
		return '<span%s%s>%s%s</span>' % (clz, data_help, inner_html, self.icon())

	# A convenient helper: returns a string with a complete <span> tag containing class and data-tt attributes,
	# without icon but with a given msg string as its inner html
	def span_noicon(self, inner_html='', clz=''):
		if clz:
			if self.css_class():
				clz='%s %s' % (clz, self.css_class())
		elif self.css_class():
			clz = self.css_class()
		if clz:
			clz = ' class="%s"' % clz
		else:
			clz=''
		if self.data_help():
			data_help = ' data-tt="%s"' % self.data_help()
		else:
			data_help = ''
		return '<span%s%s>%s</span>' % (clz, data_help, inner_html)
	
	# Returns a string containing html tags (<svg>...</svg>) displaying an icon: either error, warning or info - depending on the highest severity issue in the list
	# The icon tag contains 'data-tt' attribute with the appropriate list of issues to display tooltip
	# returns empty string if there is no issue on the list
	def icon(self, scale=1):
		icon = None
		if self.icon_override!=None:
			icon = self.icon_override
		elif self.has_error: 
			icon="icon-error"
		elif self.has_warning:
			icon="icon-warning"
		elif self.has_info: 
			icon="icon-info"
		scale = ' style="transform: scale(%.2f);"' % float(scale) if scale != 1 else ''
		if icon:
			return """ <svg class="icon" height="12px" width="12px" data-tt="%s"%s><use xlink:href="#%s"/></svg> """ % (self.data_help(), scale, icon)
		else:
			return ""
		
	# Returns a string with a name of a span. or div. class - for displaying red text for errors or orange for warnings 	
	def css_class(self):
		if self.css_class_override!=None:
			return self.css_class_override
		if self.has_error: 
			return "ERROR"
		if self.has_warning: 
			return "WARNING"
		if self.has_info: 
			return "INFO"
		return ""
	
	# Turns off showing icon if only info issues are on the list
	def turn_off_info_icon(self):
		if self.has_info and not self.has_warning and not self.has_error:
			self.icon_override = ''
	
	def override_info_css_class(self, info_css_class_override):
		if self.has_info and not self.has_warning and not self.has_error:
			self.css_class_override=info_css_class_override

	# Returns True if the list contains an issue with a given id
	def contains(self, id):
		return id in self.issues

# A series of independent routines validating a state of particular parts of a cluster. 
# Each of 'check_' function returns None if no issues were found or an 'Issues' object containing a list of found problems 
class Validator:
	def __init__(self,dp):
		self.issues_dict = []
		self.dp = dp
		# help.json is used only to determine issues severity (thus order of appearance in the list)
		# actual tooltips content is retrieved by the javascript
		with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../assets/help.json'), 'r', encoding='utf-8') as file:
			self.issues_dict = json.load(file)

	# Returns just a single issue with a given id		
	def issue(self, id):
		issues = Issues(self)
		issues.add_issue(id)
		return issues

	# Compares (stringified) versions ("4.56.3 PRO"), 
	# returns 1 if ver1>ver2, -1 if ver1<ver2, 0 if ver1==ver2, doesn't take into account "PRO" on not "PRO"
	def cmp_ver_str(self, ver1, ver2):
		(ver1_1, ver1_2, ver1_3)=list(map(int, ver1.upper().replace("PRO","").split('.')))
		(ver2_1, ver2_2, ver2_3)=list(map(int, ver2.upper().replace("PRO","").split('.')))
		if (ver1_1>ver2_1):
			return 1
		elif (ver1_1<ver2_1):
			return -1
		if (ver1_2>ver2_2):
			return 1
		elif (ver1_2<ver2_2):
			return -1
		if (ver1_3>ver2_3):
			return 1
		elif (ver1_3<ver2_3):
			return -1
		return 0

	# Checks if GUI version is newer/older than master and if it's identical in terms of PRO vs. CE
	# returns a list of issues
	def check_gui_ver(self, ver_ms, ver_gui):
		issues = Issues(self)
		ver_diff = self.cmp_ver_str(ver_ms, ver_gui)
		if ver_diff == 0:
			return issues
		if (ver_diff==1): 
			issues.add_issue('ui_older') # master in newer than GUI
		else:
			issues.add_issue('ui_newer') # GUI in newer than master
		# check PRO vs. CE mismatch
		if ("PRO" in ver_ms.upper() and not "PRO" in ver_gui.upper()):
			issues.add_issue('ui_ce_pro')
		elif (not "PRO" in  ver_ms.upper() and "PRO" in ver_gui.upper()):
			issues.add_issue('ui_pro_ce')
		return issues

	def check_ms_metadata_save_older(self, ms, highest_saved_metaversion):
		issues = Issues(self)
		if ms.lastsavemetaversion!=highest_saved_metaversion:
			issues.add_issue('ms_metadata_save_older')
		issues.turn_off_info_icon() # don't show any icon if only info on the list
		issues.override_info_css_class('GRAY') # use GRAY css class for this issue
		return issues


	def check_ms_checksum_mismatch(self, ms, highest_saved_metaversion, highest_metaversion_checksum):
		issues = Issues(self)
		if ms.lastsavemetaversion!=highest_saved_metaversion:
			issues.add_issue('ms_metadata_save_older')
		if ms.lastsavemetaversion==highest_saved_metaversion and ms.lastsavemetachecksum!=highest_metaversion_checksum:
			issues.add_issue('ms_checksum_mismatch')
		issues.turn_off_info_icon() # don't show any icon if only info on the list
		issues.override_info_css_class('GRAY') # use GRAY css class for this issue
		return issues
	
	def check_ms_last_store_status(self,ms):
		issues = Issues(self)
		if ms.lastsavestatus==LASTSTORE_META_STORED_BG:
			issues.add_issue('ms_last_store_background')
		elif ms.lastsavestatus==LASTSTORE_DOWNLOADED:
			issues.add_issue('ms_last_store_downloaded')
		elif ms.lastsavestatus==LASTSTORE_META_STORED_FG:
			issues.add_issue('ms_last_store_foreground')
		elif ms.lastsavestatus==LASTSTORE_CRC_STORED_BG:
			issues.add_issue('ms_last_store_crc_background')
			issues.css_class_override = 'NOTICE'
		else:
			issues.add_issue('ms_last_store_unknown')
		issues.turn_off_info_icon() # don't show any icon if only info on the list
		return issues
	
	# Check if follower is older or newer and if it matches PRO/CE with the Leader
	def check_follower_version(self, ms):
		issues = Issues(self)
		if self.dp.master()==None or not ms.strver:
			return issues
		# check PRO vs. CE mismatch
		if self.dp.master().is_pro() and not ms.pro:
			issues.add_issue('follower_ce_pro')
		elif not self.dp.master().is_pro() and ms.pro:
			issues.add_issue('follower_pro_ce')
		if  self.dp.master().sortver > ms.sortver:
			issues.add_issue('follower_older')
		elif self.dp.master().sortver < ms.sortver:
			issues.add_issue('follower_newer')
		return issues
	
	# Check if metalogger is older or newer and if it matches PRO/CE with the Leader
	def check_ml_version(self, ml):
		issues = Issues(self)
		if self.dp.master()==None or not ml.strver:
			return issues
		# check PRO vs. CE mismatch
		if self.dp.master().is_pro() and not ml.pro:
			issues.add_issue('ml_ce_pro')
		elif not self.dp.master().is_pro() and ml.pro:
			issues.add_issue('ml_pro_ce')
		if  self.dp.master().sortver > ml.sortver:
			issues.add_issue('ml_older')
		elif self.dp.master().sortver < ml.sortver:
			issues.add_issue('ml_newer')
		return issues
	
	# Check if chunkserver is older or newer and if it matches PRO/CE with the Leader
	def check_cs_version(self, cs):
		issues = Issues(self)
		if self.dp.master()==None or not cs.strver:
			return issues
		# check PRO vs. CE mismatch
		if self.dp.master().is_pro() and not cs.pro:
			issues.add_issue('cs_ce_pro')
		elif not self.dp.master().is_pro() and cs.pro:
			issues.add_issue('cs_pro_ce')
		if  self.dp.master().sortver > cs.sortver:
			issues.add_issue('cs_older')
		elif self.dp.master().sortver < cs.sortver:
			issues.add_issue('cs_newer')
		return issues

	# Check if mount is older or newer and if it matches PRO/CE with the Leader
	def check_mount_version(self, mnt):
		issues = Issues(self)
		if self.dp.master()==None or not mnt.strver:
			return issues
		# check PRO vs. CE mismatch
		if self.dp.master().is_pro() and not mnt.pro:
			issues.add_issue('mnt_ce_pro')
		elif not self.dp.master().is_pro() and mnt.pro:
			issues.add_issue('mnt_pro_ce')
		if  self.dp.master().sortver > mnt.sortver:
			issues.add_issue('mnt_older')
		elif self.dp.master().sortver < mnt.sortver:
			issues.add_issue('mnt_newer')
		return issues
	
	# Check if there is enough master servers
	def check_cluster_masters(self, masterservers, num_metaloggers):
		issues = Issues(self)
		active = leader = elect = usurper = 0
		pro = False
		for ms in masterservers:
			if ms.is_active():
				active+=1
				pro = ms.pro or pro
				if ms.workingstate==STATE_LEADER:
					leader += 1
				elif ms.workingstate==STATE_ELECT:
					elect += 1
				elif ms.workingstate==STATE_USURPER:
					usurper += 1
		if active==1 and pro:
			issues.add_issue('ms_single_master_pro')
		if active==1 and num_metaloggers==0:
			issues.add_issue('ms_single')
		if leader==0:
			if elect==1:
				issues.add_issue('ms_elect')
			elif usurper==1:
				issues.add_issue('ms_usurper')
		if leader+elect+usurper>1 and not (leader==1 and usurper==1 and elect==0):
			issues.add_issue('ms_toomany')
		return issues
	
	# Check (explain) the state of the master server (desync, unreachable etc.)
	def check_ms_state(self, ms):
		issues = Issues(self)
		if not ms.is_active():
			issues.add_issue('follower_unreachable')
		if ms.workingstate==STATE_FOLLOWER and "DESYNC" in ms.statestr:
			issues.add_issue('follower_desync')
		return issues
	
	# Is the follower clock in sync with the leader?
	def check_follower_clock(self,secdelta):
		issues = Issues(self)
		if secdelta>2.0:
			issues.add_issue('follower_time_diff')
		elif secdelta>1.0:
			issues.add_issue('follower_time_diff_small')
		return issues
	
	# Is the follower's metaid the same as the leader's?
	def check_follower_metaid(self, ms):
		issues = Issues(self)
		if ms.metaid!=self.dp.cluster.master_metaid():
			issues.add_issue('follower_id_mismatch')
		return issues

	# Does follower checksum match the leader's?
	def check_follower_checksum(self, ms):
		issues = Issues(self)
		if ms.exportschecksum!=self.dp.cluster.master_exportschecksum():
			issues.add_issue('follower_checksum_mismatch')
		return issues

	# What is the follower's dalay in metadata sync with the leader?
	def check_follower_metadelay(self, metadelay):
		issues = Issues(self)
		if metadelay>6.0:
			issues.add_issue('follower_meta_delay')
		elif metadelay>1.0:
			issues.add_issue('follower_meta_delay_small')
		return issues

	# Isn't the last metadata save duration too long?
	def check_last_save_duration(self, ms):
		issues = Issues(self)
		if ms.lastsaveseconds>45*60:
			issues.add_issue('ms_long_duration')
		return issues

	# Check licence expiration time
	def check_licence_time(self, lic):
		issues = Issues(self)
		if lic.licleft:
			if lic.licleft <= 0:
				issues.add_issue('lic_expired')
			elif lic.licleft < 86400 * 30:
				issues.add_issue('lic_expiring')
		return issues
	
	# Check licence size
	def check_licence_size(self, lic):
		issues = Issues(self)
		if lic.licver==0 or lic.licver == LICVER_CE: return issues
		if (lic.currentsize >= lic.licmaxsize):
			issues.add_issue('lic_exhausted')
		elif (lic.currentsize >= lic.licmaxsize*0.95):
			issues.add_issue('lic_almost_exhausted')
		elif (lic.currentsize >= lic.licmaxsize*0.80):
			issues.add_issue('lic_exhausting')
		return issues
	
	# Check version granted
	def check_licence_version(self, lic, ver):
		issues = Issues(self)
		if not lic.is_allowed_version(ver):
			issues.add_issue('lic_max_version')
		return issues
	
	# Check if the chunk matrix numbers are valid
	def check_mx_progress(self, progressstatus):
		issues = Issues(self)
		if progressstatus>0:
			issues.add_issue('data_health_checking')
		return issues

	# Check if there are any missing, endangered or undergoal chunks
	def check_mx_summary(self, mx_summary):
		issues = Issues(self)
		if mx_summary[MX_COL_MISSING]>0:
			issues.add_issue('data_health_missing')
		elif mx_summary[MX_COL_ENDANGERED]>0:
			issues.add_issue('data_health_endangered')
		elif mx_summary[MX_COL_UNDERGOAL]>0:
			issues.add_issue('data_health_undergoal')
		return issues
	
	# Check if there are issues reported by the health self-check loop
	def check_health_selfcheck(self, hsc):
		issues = Issues(self)
		if hsc.mfiles>0 or hsc.mchunks>0:
			issues.add_issue('data_missing_files')
		if hsc.ugfiles>0 or hsc.ugchunks>0:
			issues.add_issue('data_undergoal_files')
		if hsc.mtfiles>0:
			issues.add_issue('data_missing_trash_files')
		return issues
	
	# Return info on a missing chunk status
	def get_missing_chunk_info(self, mc):
		issues = Issues(self)
		if mc.mtype == MISSING_CHUNK_TYPE_NOCOPY:
			issues.add_issue('data_mc_nocopy')
		elif mc.mtype == MISSING_CHUNK_TYPE_INVALID_COPIES:
			issues.add_issue('data_mc_invalid')
		elif mc.mtype == MISSING_CHUNK_TYPE_WRONG_VERSIONS:
			issues.add_issue('data_mc_wrongver')
		elif mc.mtype == MISSING_CHUNK_TYPE_PARTIAL_EC:
			issues.add_issue('data_mc_partialec')
		return issues

	# Check CS queue overload state
	def check_cs_queue_state(self,cs):
		issues = Issues(self)
		if cs.queue_state==CS_LOAD_OVERLOADED:
			issues.add_issue('cs_overload')
		return issues
	
	def check_cs_hdds(self,cs):
		issues = Issues(self)
		if cs.total>0 and cs.used >= cs.total*0.99:
			issues.add_issue('cs_hdds_full')
		elif cs.total>0 and cs.used >= cs.total*0.90:
			issues.add_issue('cs_hdds_nearly_full')
		return issues

	def check_cs_mfr_status(self, cs):
		issues = Issues(self)
		if cs.tdchunks!=0:
			if cs.mfrstatus==MFRSTATUS_VALIDATING:
				issues.add_issue('cs_mfr_validating')
			elif cs.mfrstatus==MFRSTATUS_INPROGRESS:
				issues.add_issue('cs_mfr_inprogress')
			elif cs.mfrstatus==MFRSTATUS_READY:
				issues.add_issue('cs_mfr_ready')
			else:
				issues.add_issue('cs_mfr_unknown')
			issues.turn_off_info_icon()
		return issues

	# Check all possible hdd status issues
	def check_hdd_status(self, hdd):
		issues = Issues(self)
		if (hdd.csvalid == CS_HDD_CS_TOO_OLD):
			issues.add_issue('hdd_cs_too_old')
		elif (hdd.csvalid == CS_HDD_CS_UNREACHABLE):
			issues.add_issue('hdd_cs_unreachable')
		else:
			if (hdd.flags&CS_HDD_INVALID):
				issues.add_issue('hdd_invalid')
			if (hdd.flags&CS_HDD_DAMAGED) and (hdd.flags&CS_HDD_SCANNING)==0 and (hdd.flags&CS_HDD_INVALID)==0:
				issues.add_issue('hdd_damage')
			if hdd.flags&CS_HDD_SCANNING:
				issues.add_issue('hdd_scanning')

			if hdd.flags&CS_HDD_MFR:
				if hdd.mfrstatus==MFRSTATUS_VALIDATING:
					issues.add_issue('hdd_mfr_validating')
				elif hdd.mfrstatus==MFRSTATUS_INPROGRESS:
					issues.add_issue('hdd_mfr_inprogress')
				elif hdd.mfrstatus==MFRSTATUS_READY:
					issues.add_issue('hdd_mfr_ready')
				else:
					issues.add_issue('hdd_mfr_unknown')
				issues.turn_off_info_icon()
			if not hdd.flags&CS_HDD_DAMAGED and (hdd.errtime!=0 or hdd.errchunkid!=0):
				issues.add_issue('hdd_error')
		return issues

	# Check hdd usage (capacity) issues
	def check_hdd_usage(self, hdd):
		issues = Issues(self)
		if (hdd.csvalid != CS_HDD_CS_TOO_OLD) and (hdd.csvalid != CS_HDD_CS_UNREACHABLE):
			if hdd.total>0 and hdd.used >= hdd.total*0.99:
				issues.add_issue('hdd_full')
			elif hdd.total>0 and hdd.used >= hdd.total*0.90:
				issues.add_issue('hdd_nearly_full')
		return issues

	# Check if defining min trash retention is necessary
	def check_sc_mintrashretention(self, sc):
		issues = Issues(self)
		if not sc.defined_trash:
			if sc.min_trashretention>0:
				issues.add_issue('sc_mintrashretention_unnecessary')
			elif sc.min_trashretention==0:
				issues.add_issue('sc_mintrashretention_na')
		else:
			if sc.min_trashretention==0:
				issues.add_issue('sc_mintrashretention0')
		return issues



import json

from common.constants import *
from common.utils import *

#######################################
# A set of classes representing data models for particular entities in a cluster (servers, disks, sessions, storage classes etc.)
class MasterServer:
	def __init__(self,host,port,versionxyzp,workingstate,sync,statestr,statecolor,metaversion,memusage,syscpu,usercpu,lastsuccessfulstore,lastsaveseconds,lastsavestatus,exportschecksum,metaid,lastsavemetaversion,lastsavemetachecksum,usectime,chlogtime):
		self.host = host
		self.port = port
		try:
			iptab = tuple(map(int,host.split('.')))
			self.strip = "%u.%u.%u.%u" % iptab
			self.sortip = "%03u_%03u_%03u_%03u" % iptab
		except Exception:
			self.strip = host
			self.sortip = host
		self.version = versionxyzp
		self.strver, self.sortver, self.pro = version_str_sort_pro(versionxyzp)
		self.workingstate = workingstate
		self.sync = sync
		self.statestr = statestr
		self.statecolor = statecolor
		self.metaversion = metaversion
		self.memusage = memusage
		self.syscpu = syscpu
		self.usercpu = usercpu
		self.lastsuccessfulstore = lastsuccessfulstore
		self.lastsaveseconds = lastsaveseconds
		self.lastsavestatus = lastsavestatus
		self.exportschecksum = exportschecksum
		self.metaid = metaid
		self.lastsavemetaversion = lastsavemetaversion
		self.lastsavemetachecksum = lastsavemetachecksum
		self.usectime = usectime
		self.chlogtime = chlogtime
		self.secdelta = None  #updated separately by update_masterservers_delays
		self.metadelay = None #updated separately by update_masterservers_delays
		
		if self.workingstate==STATE_FOLLOWER:
			if sync==0:	self.statestr += " (DESYNC)"
			if sync==2:	self.statestr += " (DELAYED)"
			if sync==3:	self.statestr += " (INIT)"

	def is_active(self):
		return self.workingstate != STATE_UNREACHABLE

# 		self.featuremask = 0
# 		if self.version>=(3,0,72):
# 			self.featuremask |= (1<<FEATURE_EXPORT_UMASK)
# 		if (self.version>=(3,0,112) and self.version[0]==3) or self.version>=(4,21,0):
# 			self.featuremask |= (1<<FEATURE_EXPORT_DISABLES)
# 		if self.version>=(4,27,0):
# 			self.featuremask |= (1<<FEATURE_SESSION_STATS_28)
# 		if self.version>=(4,29,0):
# 			self.featuremask |= (1<<FEATURE_INSTANCE_NAME)
# 		if self.version>=(4,35,0):
# 			self.featuremask |= (1<<FEATURE_CSLIST_MODE)
# 		if self.version>=(4,44,0):
# 			self.featuremask |= (1<<FEATURE_SCLASS_IN_MATRIX)
# 		if self.version>=(4,51,0):
# 			self.featuremask |= (1<<FEATURE_DEFAULT_GRACEPERIOD)
# 		if self.version>=(4,53,0):
# 			self.featuremask |= (1<<FEATURE_LABELMODE_OVERRIDES)
# 		if self.version>=(4,57,0):
# 			self.featuremask |= (1<<FEATURE_SCLASSGROUPS)	

# 	def version_at_least(self,v1,v2,v3):
# 		return (self.version>=(v1,v2,v3))
# 	def version_less_than(self,v1,v2,v3):
# 		return (self.version<(v1,v2,v3))
# 	def version_is(self,v1,v2,v3):
# 		return (self.version==(v1,v2,v3))
# 	def version_unknown(self):
# 		return (self.version==(0,0,0))
# 	def is_pro(self):
# 		return self.pro
# 	def has_feature(self,featureid):
# 		return True if (self.featuremask & (1<<featureid)) else False
	


class Metalogger:
	# ip should be tuple: (ip1,ip2,ip3,ip4), version should be tuple: (v1,v2,v3)
	def __init__(self, ip, donotresolve, version):
		self.strip = "%u.%u.%u.%u" % ip
		self.sortip = "%03u_%03u_%03u_%03u" % ip
		self.host = resolve(self.strip, donotresolve)
		self.strver,self.sortver,self.pro = version_str_sort_pro(version)

class ClusterInfo:
	def __init__(self, version,memusage,syscpu,usercpu,totalspace,availspace,freespace,trspace,trfiles,respace,refiles,nodes,dirs,files,chunks,lastsuccessfulstore,lastsaveseconds,lastsavestatus,metainfomode,
																	allcopies,regularcopies,copychunks,ec8chunks,ec4chunks,chunkcopies,chunkec8parts,chunkec4parts,chunkhypcopies):
		self.version = version
		self.memusage = memusage
		self.syscpu = syscpu
		self.usercpu = usercpu
		self.totalspace = totalspace
		self.availspace = availspace
		self.freespace = freespace
		self.trspace = trspace
		self.trfiles = trfiles
		self.respace = respace
		self.refiles = refiles
		self.nodes = nodes
		self.dirs = dirs
		self.files = files
		self.chunks = chunks
		self.lastsuccessfulstore = lastsuccessfulstore
		self.lastsaveseconds = lastsaveseconds
		self.lastsavestatus = lastsavestatus
		self.metainfomode = metainfomode
		self.allcopies = allcopies
		self.regularcopies = regularcopies
		self.copychunks = copychunks
		self.ec8chunks = ec8chunks
		self.ec4chunks = ec4chunks
		self.chunkcopies = chunkcopies
		self.chunkec8parts = chunkec8parts
		self.chunkec4parts = chunkec4parts
		self.chunkhypcopies = chunkhypcopies
		self.strver,self.sortver,_ = version_str_sort_pro(version)

		if metainfomode==1:
			if (copychunks + ec8chunks + ec4chunks) > 0:
				self.ecchunkspercent = 100.0 * (ec8chunks + ec4chunks) / (copychunks + ec8chunks + ec4chunks)
				self.dataredundancyratio = (chunkcopies + 0.125 * chunkec8parts + 0.25 * chunkec4parts) / (copychunks + ec8chunks + ec4chunks)
			else:
				self.ecchunkspercent = None
				self.dataredundancyratio = None
			if (chunkcopies + 0.125 * chunkec8parts + 0.25 * chunkec4parts) > 0.0:
				noecratio = (chunkhypcopies / (chunkcopies + 0.125 * chunkec8parts + 0.25 * chunkec4parts))
				if noecratio>1.0:
					extranoecspace = noecratio - 1.0
				else:
					extranoecspace = 0.0
				if freespace!=None:
					self.savedbyec = int((totalspace-freespace) * extranoecspace)
				else:
					self.savedbyec = int((totalspace-availspace) * extranoecspace)
			else:
				self.savedbyec = None



class Licence:
	def __init__(self,licver,lictype,v1,v2,v3,licmaxtime,licleft,licmaxsize,currentsize,licid,lictypestr,licissuer,licuser,licjson,addinfo,quota,csmaxsize,csmaxcnt):
		self.licver = licver
		self.lictype = lictype
		self.v1 = v1
		self.v2 = v2
		self.v3 = v3
		self.licmaxtime = licmaxtime
		self.licleft = licleft
		self.licmaxsize = licmaxsize
		self.currentsize = currentsize
		self.licid = licid
		self.lictypestr = lictypestr
		self.licissuer = licissuer
		self.licuser = licuser
		self.licjson = licjson
		self.addinfo = addinfo
		self.quota = quota
		self.csmaxsize = csmaxsize
		self.csmaxcnt = csmaxcnt
		if self.licver and self.licver>=1 and len(self.licjson)>0:
			self.licextrainfo = json.loads(self.licjson)
		else:
			self.licextrainfo = {}

	def get_max_version_str(self):
		if self.v1==0xFFFF or self.licver == LICVER_CE:
			return "unlimited"
		elif self.v2==255:
			return "%u.x" % (self.v1)
		elif self.v3==255:
			return "%u.%u.x" % (self.v1,self.v2)
		else:
			return "%u.%u.%u" % (self.v1,self.v2,self.v3)

	# Returns true if given software version (v1,v2,v3) is allowed by the licence
	def is_allowed_version(self,version):
		if (self.licver == LICVER_CE):
			return True
		v1,v2,v3 = version
		if self.v1==0xFFFF:
			return True #unlimited
		if v1<self.v1:
			return True
		elif v1>self.v1:
			return False
		if self.v2==255:
			True		#A.x unlimited
		if v2<self.v2:
			return True
		elif v2>self.v2:
			return False
		if self.v3==255:
			True		#A.B.x unlimited
		if v3<self.v3:
			return True
		elif v3>self.v3:
			return False
		return True

	def is_time_unlimited(self):
		return self.licleft==0xFFFFFFFF or self.licmaxtime==0xFFFFFFFF or self.licver == LICVER_CE
	
	def is_cs_size_unlimited(self):
		return self.csmaxsize==0xFFFFFFFFFFFFFFFF or self.licver == LICVER_CE
	
	def is_cs_cnt_unlimited(self):
		return self.csmaxcnt==0xFFFF or self.licver == LICVER_CE
	
	def is_size_unlimited(self):
		return self.licmaxsize==0xFFFFFFFFFFFFFFFF or (self.licmaxsize==0 and self.currentsize==0) or self.licver == LICVER_CE


class HealthSelfCheck:
	def __init__(self,loopstart,loopend,files,ugfiles,mfiles,mtfiles,msfiles,chunks,ugchunks,mchunks,msgbuffleng,datastr):
		self.loopstart = loopstart
		self.loopend = loopend
		self.files = files
		self.ugfiles = ugfiles
		self.mfiles = mfiles
		self.mtfiles = mtfiles
		self.msfiles = msfiles
		self.chunks = chunks
		self.ugchunks = ugchunks
		self.mchunks = mchunks
		self.msgbuffleng = msgbuffleng
		self.datastr = datastr

# Missing chunk
class MissingChunk:
	def __init__(self,paths,inode,indx,chunkid,mtype):
		self.paths = paths
		self.inode = inode
		self.indx = indx
		self.chunkid = chunkid
		self.mtype = mtype

	def get_mtype_str(self):
		if self.mtype==MISSING_CHUNK_TYPE_NOCOPY:
			return "No copy"
		elif self.mtype==MISSING_CHUNK_TYPE_INVALID_COPIES:
			return "Invalid copies"
		elif self.mtype==MISSING_CHUNK_TYPE_WRONG_VERSIONS:
			return "Wrong versions"
		elif self.mtype==MISSING_CHUNK_TYPE_PARTIAL_EC:
			return "Missing EC parts"
		else:
			return "Other"

	def get_chunkid_str(self):
		return "%016X" % self.chunkid

class MemoryUsage:
	def __init__(self, memlabels,abrlabels,totalused,totalallocated,memusage):
		self.memlabels = memlabels
		self.abrlabels = abrlabels
		self.totalused = totalused
		self.totalallocated = totalallocated
		self.memusage = memusage

class ChunkServer:
	def __init__(self,oip,ip,donotresolve,port,csid,v1,v2,v3,flags,used,total,chunks,tdused,tdtotal,tdchunks,errcnt,queue,gracetime,labels,mfrstatus,maintenanceto):
		self.oip = oip
		self.ip = ip
		self.version = (v1,v2,v3)
		self.stroip = "%u.%u.%u.%u" % oip
		self.strip = "%u.%u.%u.%u" % ip
		self.sortip = "%03u_%03u_%03u_%03u" % ip
		self.strver,self.sortver,self.pro = version_str_sort_pro((v1,v2,v3))
		self.host = resolve(self.strip, donotresolve)
		self.port = port
		self.csid = csid
		self.flags = flags
		self.used = used
		self.total = total
		self.chunks = chunks
		self.tdused = tdused
		self.tdtotal = tdtotal
		self.tdchunks = tdchunks
		self.errcnt = errcnt
		self.queue = queue
		self.gracetime = gracetime
		self.labels = labels
		self.mfrstatus = mfrstatus
		self.maintenanceto = maintenanceto
		self.hostkey = "%s:%s" % (self.strip,self.port)

		if self.is_maintenance_off():
			self.mmto = "n/a"
		elif self.maintenanceto==None:
			self.mmto = "unknown"
		elif self.maintenanceto==0xFFFFFFFF:
			self.mmto = "permanent"
		else:
			self.mmto = "%u seconds left" % self.maintenanceto

		if self.gracetime>=0xC0000000:
			self.queue_state = CS_LOAD_OVERLOADED
			self.queue_state_str = "Overloaded"
			self.queue_state_msg = "Overloaded"
			self.queue_state_info = "server queue is heavy loaded (overloaded)"
			self.queue_cgi = "<%u>" % self.queue
		elif self.gracetime>=0x80000000:
			self.queue_state = CS_LOAD_REBALANCE
			self.queue_state_str = "Rebalance"
			self.queue_state_msg = "Rebalancing"
			self.queue_state_info = "internal rebalance in progress"
			self.queue_cgi = "(%u)" % self.queue
		elif self.gracetime>=0x40000000:
			self.queue_state = CS_LOAD_FAST_REBALANCE
			self.queue_state_str = "Fast rebalance"
			self.queue_state_msg = "Fast rebalancing"
			self.queue_state_info = "high speed rebalance in progress"
			self.queue_cgi = "{%u}" % self.queue
			# queue_state_has_link = 1
		elif self.gracetime>0:
			self.queue_state = CS_LOAD_GRACEFUL
			self.queue_state_str = "G(%u)" % self.gracetime
			self.queue_state_msg = "%u secs graceful" % self.gracetime
			self.queue_state_info = "server in graceful period - back to normal after %u seconds" % self.gracetime
			self.queue_cgi = "[%u]" % self.queue
		else:
			self.queue_state = CS_LOAD_NORMAL
			self.queue_state_str = "Normal"
			self.queue_state_msg = "Normal load"
			self.queue_state_info = "server is responsive, working with low load"
			self.queue_cgi = "%u" % self.queue

		if self.labels==0xFFFFFFFF or self.labels==0:
			self.labelstr = "-"
		else:
			labelstab = []
			for bit,char in enumerate(map(chr,range(ord('A'),ord('Z')+1))):
				if self.labels & (1<<bit):
					labelstab.append(char)
			self.labelstr = ",".join(labelstab)		

	def is_maintenance_off(self):
		return (self.flags&6)==0

	def is_maintenance_on(self):
		return (self.flags&6)==2

	def is_connected(self):
		return (self.flags&1)==0

	def get_mfr_status_str(self):
		if self.tdchunks!=0:
			if self.mfrstatus==MFRSTATUS_VALIDATING:
				return 'not ready for removal (validating)'
			elif self.mfrstatus==MFRSTATUS_INPROGRESS:
				return 'not ready for removal (in progress)'
			elif self.mfrstatus==MFRSTATUS_READY:
				return 'removal ready'
			else:
				return 'not ready for removal (unrecognized status)'
		else:
			return ''
		
# Chunkserver's hard disk drive
class HDD:
	def __init__(self,csvalid,hostkey,hoststr,hostip,port,hddpath,sortippath,ippath,hostpath,flags,clearerrorarg,errchunkid,errtime,used,total,chunkscnt,rbw,wbw,usecreadavg,usecwriteavg,usecfsyncavg,usecreadmax,usecwritemax,usecfsyncmax,rops,wops,fsyncops,rbytes,wbytes,mfrstatus):
		self.csvalid = csvalid
		self.hostkey = hostkey
		self.hoststr = hoststr
		self.hostip = hostip
		self.port = port
		self.hddpath = hddpath
		self.sortippath = sortippath
		self.ippath = ippath
		self.hostpath = hostpath
		self.flags = flags
		self.clearerrorarg = clearerrorarg
		self.errchunkid = errchunkid
		self.errtime = errtime
		self.used = used
		self.total = total
		self.chunkscnt = chunkscnt
		self.rbw = rbw
		self.wbw = wbw
		self.usecreadavg = usecreadavg
		self.usecwriteavg = usecwriteavg
		self.usecfsyncavg = usecfsyncavg
		self.usecreadmax = usecreadmax
		self.usecwritemax = usecwritemax
		self.usecfsyncmax = usecfsyncmax
		self.rops = rops
		self.wops = wops
		self.fsyncops = fsyncops
		self.rbytes = rbytes
		self.wbytes = wbytes
		self.mfrstatus = mfrstatus

	def __lt__(self, other):
		return self.sortippath < other.sortippath
	
	def is_valid(self):
		return self.csvalid==CS_HDD_CS_VALID
	def has_errors(self):
		return self.errtime!=0 or self.errchunkid!=0
	
	def get_statuslist(self):
		statuslist = []
		if (self.csvalid == CS_HDD_CS_TOO_OLD):
			statuslist.append('unknown (too old chunkserver)')
		elif (self.csvalid == CS_HDD_CS_UNREACHABLE):
			statuslist.append('unknown (unreachable chunkserver)')
		else:
			if (self.flags&CS_HDD_INVALID):
				statuslist.append('invalid')
			if (self.flags&CS_HDD_DAMAGED) and (self.flags&CS_HDD_SCANNING)==0 and (self.flags&CS_HDD_INVALID)==0:
				statuslist.append('damaged')
			if self.flags&CS_HDD_MFR:
				if self.mfrstatus==MFRSTATUS_VALIDATING:
					statuslist.append('removal not ready (validating)')
				elif self.mfrstatus==MFRSTATUS_INPROGRESS:
					statuslist.append('removal not ready (in progress)')
				elif self.mfrstatus==MFRSTATUS_READY:
					statuslist.append('removal ready')
				else:
					statuslist.append('removal not ready (unrecognized status)')
			if self.flags&CS_HDD_SCANNING:
				statuslist.append('scanning')
			if self.flags==0:
				if self.has_errors():
					statuslist.append('ok (errors)')
				else:
					statuslist.append('ok')
		return statuslist

	# Returns a string with a status of the HDD
	def get_status_str(self):
		return ", ".join(self.get_statuslist())

class ExportsEntry:
	def __init__(self,fip1,fip2,fip3,fip4,tip1,tip2,tip3,tip4,path,meta,v1,v2,v3,exportflags,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mingoal,maxgoal,mintrashretention,maxtrashretention,disables):
		self.ipfrom = (fip1,fip2,fip3,fip4)
		self.ipto = (tip1,tip2,tip3,tip4)
		self.version = (v1,v2,v3)
		self.stripfrom = "%u.%u.%u.%u" % (fip1,fip2,fip3,fip4)
		self.sortipfrom = "%03u_%03u_%03u_%03u" % (fip1,fip2,fip3,fip4)
		self.stripto = "%u.%u.%u.%u" % (tip1,tip2,tip3,tip4)
		self.sortipto = "%03u_%03u_%03u_%03u" % (tip1,tip2,tip3,tip4)
		self.strver,self.sortver,self.pro = version_str_sort_pro((v1,v2,v3))
		self.meta = meta
		self.path = path
		self.exportflags = exportflags
		self.sesflags = sesflags
		self.umaskval = umaskval
		self.rootuid = rootuid
		self.rootgid = rootgid
		self.mapalluid = mapalluid
		self.mapallgid = mapallgid
		if sclassgroups==None:
			if mingoal==None and maxgoal==None:
				self.sclassgroups = 0xFFFF
			else:
				self.sclassgroups = ((0xFFFF<<mingoal) & (((0xFFFF<<(maxgoal+1))&0xFFFF)^0xFFFF)) | 1
		else:
			self.sclassgroups = sclassgroups
		self.mintrashretention = mintrashretention
		self.maxtrashretention = maxtrashretention
		self.disables = disables

		if self.sclassgroups==0:
			self.sclassgroups_sort = 0
			self.sclassgroups_str = 'NONE'
		elif self.sclassgroups==0xFFFF:
			self.sclassgroups_sort = 0xFFFF
			self.sclassgroups_str = "ALL"
		else:
			self.sclassgroups_sort = 0
			sclassgroups_list = []
			for b in range(16):
				self.sclassgroups_sort<<=1
				if self.sclassgroups & (1<<b):
					sclassgroups_list.append(b)
					self.sclassgroups_sort |= 1
			self.sclassgroups_str = ",".join(map(str,sclassgroups_list))

	def is_alldirs(self):
		return self.exportflags&1
	def is_password(self):
		return self.exportflags&2
	def is_readonly(self):
		return self.sesflags&1
	def is_unrestricted(self):
		return self.sesflags&2
	def ignore_gid(self):
		return self.sesflags&4
	def is_admin(self):
		return self.sesflags&8
	def map_user(self):
		return self.sesflags&16


class Session:
	def __init__(self,sessionid,ip,donotresolve,info,openfiles,nsocks,expire,v1,v2,v3,meta,path,sesflags,umaskval,rootuid,rootgid,mapalluid,mapallgid,sclassgroups,mingoal,maxgoal,mintrashretention,maxtrashretention,disables,stats_c,stats_l):
		self.ip = ip
		self.version = (v1,v2,v3)
		self.strip = "%u.%u.%u.%u" % ip
		self.sortip = "%03u_%03u_%03u_%03u" % ip
		self.strver,self.sortver,self.pro = version_str_sort_pro((v1,v2,v3))
		self.host = resolve(self.strip, donotresolve)
		self.sessionid = sessionid
		self.info = info
		self.openfiles = openfiles
		self.nsocks = nsocks
		self.expire = expire
		self.meta = meta
		self.path = path
		self.sesflags = sesflags
		self.umaskval = umaskval
		self.rootuid = rootuid
		self.rootgid = rootgid
		self.mapalluid = mapalluid
		self.mapallgid = mapallgid
		if sclassgroups==None:
			if mingoal==None and maxgoal==None:
				self.sclassgroups = 0xFFFF
			else:
				self.sclassgroups = ((0xFFFF<<mingoal) & (((0xFFFF<<(maxgoal+1))&0xFFFF)^0xFFFF)) | 1
		else:
			self.sclassgroups = sclassgroups
		self.mintrashretention = mintrashretention
		self.maxtrashretention = maxtrashretention
		self.disables = disables
		self.stats_c = stats_c
		self.stats_l = stats_l

	def get_sessionstr(self):
		if self.sessionid & 0x80000000:
			return "TMP/%u" % (self.sessionid&0x7FFFFFFF)
		else:
			return "%u" % (self.sessionid)
		
	def is_tmp(self):
		return self.sessionid & 0x80000000

	def get_sclassgroups_sort(self):
		if self.sclassgroups==0:
			sclassgroups_sort = 0
		elif self.sclassgroups==0xFFFF:
			sclassgroups_sort = 0xFFFF
		else:
			sclassgroups_sort = 0
			for b in range(16):
				sclassgroups_sort<<=1
				if self.sclassgroups & (1<<b):
					sclassgroups_sort |= 1
		return sclassgroups_sort

	def get_sclassgroups_str(self):
		if self.sclassgroups==0:
			sclassgroups_str = 'NONE'
		elif self.sclassgroups==0xFFFF:
			sclassgroups_str = "ALL"
		else:
			sclassgroups_list = []
			for b in range(16):
				if self.sclassgroups & (1<<b):
					sclassgroups_list.append(b)
			sclassgroups_str = ",".join(map(str,sclassgroups_list))
		return sclassgroups_str

# Represents just a short info about sclass used for sclass listing
class StorageClassName:
	def __init__(self,name,has_chunks):
		self.name = name
		self.has_chunks = has_chunks

class StorageClass:
	def __init__(self,sclassid,sclassname,sclassdesc,priority,export_group,admin_only,labels_mode,arch_mode,arch_delay,arch_min_size,min_trashretention,files,dirs,states,availableservers,labels_ver):
		self.sclassid = sclassid
		self.sclassname = sclassname
		self.sclassdesc = sclassdesc
		self.priority = priority
		self.export_group = export_group
		self.admin_only = admin_only
		self.labels_mode = labels_mode
		self.arch_mode = arch_mode
		self.arch_delay = arch_delay
		self.arch_min_size = arch_min_size
		self.min_trashretention = min_trashretention
		self.files = files
		self.dirs = dirs
		self.states = states
		self.availableservers = availableservers
		self.labels_ver = labels_ver

		arch_mode_list = []
		if arch_mode&SCLASS_ARCH_MODE_CHUNK:
			arch_mode_list.append("CHUNK")
		elif arch_mode&SCLASS_ARCH_MODE_FAST:
			arch_mode_list.append("FAST")
		else:
			arch_mode_list.append("[")
			if arch_mode&SCLASS_ARCH_MODE_CTIME:
				arch_mode_list.append("C")
			if arch_mode&SCLASS_ARCH_MODE_MTIME:
				arch_mode_list.append("M")
			if arch_mode&SCLASS_ARCH_MODE_ATIME:
				arch_mode_list.append("A")
			arch_mode_list.append("]TIME")
			if arch_mode&SCLASS_ARCH_MODE_REVERSIBLE:
				arch_mode_list.append("/REVERSIBLE")
			else:
				arch_mode_list.append("/ONEWAY")
		self.arch_mode_str = "".join(arch_mode_list)

		self.defined_create = False
		self.defined_archive = False
		self.defined_trash = False	
		for state in self.states:
			if state.name=="CREATE":
				self.defined_create = state.defined
			elif state.name=="ARCH":
				self.defined_archive = state.defined
			elif state.name=="TRASH":
				self.defined_trash = state.defined
		# admin_only_str = "YES" if admin_only else "NO"
		# labels_mode_str = "LOOSE" if labels_mode==0 else "STD" if labels_mode==1 else "STRICT"
		# arch_delay_str = hours_to_str(arch_delay) if (arch_delay>0 and (arch_mode&SCLASS_ARCH_MODE_FAST)==0) else "-"
		# min_trashretention_str = hours_to_str(min_trashretention) if min_trashretention>0 else "-"

	def get_admin_only_str(self):
		return "YES" if self.admin_only else "NO"
	def get_labels_mode_str(self):
		return "LOOSE" if self.labels_mode==0 else "STD" if self.labels_mode==1 else "STRICT"
	def get_min_trashretention_str(self):
		return hours_to_str(self.min_trashretention) if self.min_trashretention>0 else "-"
	def get_arch_delay_str(self):
		return hours_to_str(self.arch_delay) if (self.arch_delay>0 and (self.arch_mode&SCLASS_ARCH_MODE_FAST)==0) else "-"
	def get_arch_mode_str(self):
		return self.arch_mode_str
	def isKeepOnly(self):
		return not (self.defined_create or self.defined_archive or self.defined_trash)

class StorageClassState:
	def __init__(self,defined,name,counters,ec_level,labellist,uniqmask,labelsmodeover,canbefulfilled):
		self.defined = defined
		self.name = name
		self.counters = counters
		self.ec_level = ec_level
		self.labellist = labellist
		self.uniqmask = uniqmask
		self.labelsmodeover = labelsmodeover
		self.canbefulfilled = canbefulfilled
		self.fullname = "ARCHIVE" if name=="ARCH" else name

		self.ec_data_parts = 0
		self.ec_chksum_parts = 0
		if self.ec_level!=None and self.ec_level>0:
			self.ec_data_parts = self.ec_level >> 4
			self.ec_chksum_parts = self.ec_level & 0xF
			if self.ec_data_parts==0:
				self.ec_data_parts = 8

	def isEC(self):
		return self.ec_level!=None and self.ec_level>0

class OPattern:
	def __init__(self,globname,euid,egid,priority,sclassname,trashretention,seteattr,clreattr):
		self.globname = globname
		self.euid = euid
		self.egid = egid
		self.priority = priority
		self.sclassname = sclassname
		self.trashretention = trashretention
		self.seteattr = seteattr
		self.clreattr = clreattr

		self.euidstr = "ANY" if self.euid==PATTERN_EUGID_ANY else ("%u" % self.euid)
		self.egidstr = "ANY" if self.egid==PATTERN_EUGID_ANY else ("%u" % self.egid)
		self.eattrstr = eattr_to_str(self.seteattr,self.clreattr)

class ChunkTestInfo72:
	def __init__(self,loopstart,loopend,del_invalid,ndel_invalid,del_unused,ndel_unused,del_dclean,ndel_dclean,del_ogoal,ndel_ogoal,rep_ugoal,nrep_ugoal,rep_wlab,nrep_wlab,rebalance,labels_dont_match,locked_unused,locked_used):
		self.loopstart = loopstart
		self.loopend = loopend
		self.del_invalid = del_invalid
		self.ndel_invalid = ndel_invalid
		self.del_unused = del_unused
		self.ndel_unused = ndel_unused
		self.del_dclean = del_dclean
		self.ndel_dclean = ndel_dclean
		self.del_ogoal = del_ogoal
		self.ndel_ogoal = ndel_ogoal
		self.rep_ugoal = rep_ugoal
		self.nrep_ugoal = nrep_ugoal
		self.rep_wlab = rep_wlab
		self.nrep_wlab = nrep_wlab
		self.rebalance = rebalance
		self.labels_dont_match = labels_dont_match
		self.locked_unused = locked_unused
		self.locked_used = locked_used		

class ChunkTestInfo96:
	def __init__(self,loopstart,loopend,fixed,forcekeep,delete_invalid,delete_no_longer_needed,delete_wrong_version,delete_duplicated_ecpart,delete_excess_ecpart,delete_excess_copy,delete_diskclean_ecpart,delete_diskclean_copy,replicate_dupserver_ecpart,replicate_needed_ecpart,replicate_needed_copy,replicate_wronglabels_ecpart,replicate_wronglabels_copy,split_copy_into_ecparts,join_ecparts_into_copy,recover_ecpart,calculate_ecchksum,locked_unused,locked_used,replicate_rebalance):
		self.loopstart = loopstart
		self.loopend = loopend
		self.fixed = fixed
		self.forcekeep = forcekeep
		self.delete_invalid = delete_invalid
		self.delete_no_longer_needed = delete_no_longer_needed
		self.delete_wrong_version = delete_wrong_version
		self.delete_duplicated_ecpart = delete_duplicated_ecpart
		self.delete_excess_ecpart = delete_excess_ecpart
		self.delete_excess_copy = delete_excess_copy
		self.delete_diskclean_ecpart = delete_diskclean_ecpart
		self.delete_diskclean_copy = delete_diskclean_copy
		self.replicate_dupserver_ecpart = replicate_dupserver_ecpart
		self.replicate_needed_ecpart = replicate_needed_ecpart
		self.replicate_needed_copy = replicate_needed_copy
		self.replicate_wronglabels_ecpart = replicate_wronglabels_ecpart
		self.replicate_wronglabels_copy = replicate_wronglabels_copy
		self.split_copy_into_ecparts = split_copy_into_ecparts
		self.join_ecparts_into_copy = join_ecparts_into_copy
		self.recover_ecpart = recover_ecpart
		self.calculate_ecchksum = calculate_ecchksum
		self.locked_unused = locked_unused
		self.locked_used = locked_used
		self.replicate_rebalance = replicate_rebalance
	
class OpenFile:
	def __init__(self,sessionid,host,sortipnum,ipnum,info,inode,paths):
		self.sessionid = sessionid
		self.host = host
		self.sortipnum = sortipnum
		self.ipnum = ipnum
		self.info = info
		self.inode = inode
		self.paths = paths

class AcquiredLock:
	def __init__(self,inode,sessionid,host,sortipnum,ipnum,info,owner,pid,start,end,ctype):
		self.inode = inode
		self.sessionid = sessionid
		self.host = host
		self.sortipnum = sortipnum
		self.ipnum = ipnum
		self.info = info
		self.owner = owner
		self.pid = pid
		self.start = start
		self.end = end
		self.ctype = ctype

		if self.pid==0 and self.start==0 and self.end==0:
			self.locktype = "FLOCK"
		else:
			self.locktype = "POSIX"

class Quota:
	def __init__(self,path,exceeded,qflags,graceperiod,timetoblock,sinodes,slength,ssize,srealsize,hinodes,hlength,hsize,hrealsize,cinodes,clength,csize,crealsize):
		self.path = path
		self.exceeded = exceeded
		self.qflags = qflags
		self.graceperiod = graceperiod
		self.timetoblock = timetoblock
		self.sinodes = sinodes
		self.slength = slength
		self.ssize = ssize
		self.srealsize = srealsize
		self.hinodes = hinodes
		self.hlength = hlength
		self.hsize = hsize
		self.hrealsize = hrealsize
		self.cinodes = cinodes
		self.clength = clength
		self.csize = csize
		self.crealsize = crealsize
